# pyright: reportImportCycles=false
from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple, final

if TYPE_CHECKING:
    import asyncio
    from collections.abc import Awaitable, Callable
    from datetime import datetime

    from jobify._internal.common.datastructures import RequestState, State
    from jobify._internal.configuration import (
        JobifyConfiguration,
        RouteOptions,
    )
    from jobify._internal.inspection import FuncSpec
    from jobify._internal.message import Triggers
    from jobify._internal.runners import Runnable
    from jobify._internal.scheduler.job import Job
    from jobify._internal.scheduler.scheduler import ScheduleBuilder


@final
class OuterContext:
    __slots__: tuple[str, ...] = (
        "arguments",
        "func_spec",
        "is_force",
        "is_persist",
        "is_replace",
        "job",
        "jobify_config",
        "persist_job_hook",
        "request_state",
        "route_options",
        "runnable",
        "schedule_builder",
        "schedule_hook",
        "state",
        "trigger",
    )

    def __init__(  # noqa: PLR0913
        self,
        *,
        job: Job[Any],
        state: State,
        trigger: Triggers,
        runnable: Runnable[Any],
        arguments: dict[str, Any],
        func_spec: FuncSpec[Any],
        is_force: bool,
        is_persist: bool,
        is_replace: bool,
        route_options: RouteOptions,
        jobify_config: JobifyConfiguration,
        request_state: RequestState,
        persist_job_hook: Callable[[str, datetime, Triggers], Awaitable[None]],
        schedule_hook: Callable[[], asyncio.Handle],
        schedule_builder: ScheduleBuilder[Any],
    ) -> None:
        self.job = job
        self.state = state
        self.trigger = trigger
        self.runnable = runnable
        self.arguments = arguments
        self.func_spec = func_spec
        self.is_force = is_force
        self.is_persist = is_persist
        self.is_replace = is_replace
        self.route_options = route_options
        self.jobify_config = jobify_config
        self.request_state = request_state
        self.persist_job_hook = persist_job_hook
        self.schedule_hook = schedule_hook
        self.schedule_builder = schedule_builder


class JobContext(NamedTuple):
    job: Job[Any]
    state: State
    runnable: Runnable[Any]
    request_state: RequestState
    route_options: RouteOptions
    jobify_config: JobifyConfiguration
    schedule_builder: ScheduleBuilder[Any]


def _resolve_type_key(field_type: Any) -> str:  # noqa: ANN401
    """Extract a string key from a type annotation for matching.

    Handles both resolved types and forward references without triggering
    circular imports.
    """
    # Check if it's a ForwardRef (unresolved string annotation)
    if hasattr(field_type, "__forward_arg__"):
        # Extract base type name from forward ref
        forward_arg: str = field_type.__forward_arg__
        return forward_arg.split("[", 1)[0]
    # For resolved types, use __name__ or __qualname__
    return getattr(
        field_type,
        "__qualname__",
        getattr(field_type, "__name__", str(field_type)),
    )


def _make_type_map(tp: type[Any]) -> dict[str, str]:
    """Build a mapping from type name to field name for injection lookups."""
    return {
        _resolve_type_key(field_type): field_name
        for field_name, field_type in tp.__annotations__.items()
    }


CONTEXT_TYPE_MAP = _make_type_map(JobContext)


def inject_context(context: JobContext) -> None:
    runnable = context.runnable
    arguments = runnable.bound.arguments
    for name, tp in runnable.func_spec.inject_params.items():
        if tp is JobContext:
            val = context
        elif (
            field_name := CONTEXT_TYPE_MAP.get(_resolve_type_key(tp))
        ) is not None:
            val = getattr(context, field_name)
        else:
            msg = (
                f"Unknown type for injection: {tp}. "
                f"Available types: {list(CONTEXT_TYPE_MAP.keys())}"
            )
            raise ValueError(msg)
        arguments[name] = val



# ruff: noqa: ERA001
# pyright: reportExplicitAny=false

import pickle  # nosec B403
from typing import Any

from typing_extensions import override

from jobify._internal.serializers.base import Serializer


class UnsafePickleSerializer(Serializer):
    @override
    def dumpb(self, data: Any) -> bytes:
        # nosemgrep: python.lang.security.deserialization.pickle.avoid-pickle
        return pickle.dumps(data)

    @override
    def loadb(self, data: bytes) -> Any:
        # nosemgrep: python.lang.security.deserialization.pickle.avoid-pickle
        return pickle.loads(data)  # noqa: S301 # nosec B301



from __future__ import annotations

from typing import Any

from typing_extensions import override

from jobify._internal.typeadapter.base import Dumper, Loader


class DummyLoader(Loader):
    @override
    def load(self, data: Any, tp: Any, /) -> Any:
        return data


class DummyDumper(Dumper):
    @override
    def dump(self, data: Any, tp: Any, /) -> Any:
        return data



import re
from abc import ABCMeta, abstractmethod
from collections.abc import Sequence
from datetime import datetime
from typing import NamedTuple, Protocol

from jobify._internal.common.constants import JobStatus


class ScheduledJob(NamedTuple):
    job_id: str
    name: str
    message: bytes
    status: JobStatus
    next_run_at: datetime


def validate_table_name(table_name: str) -> None:
    if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", table_name):
        msg = (
            f"Invalid table name: {table_name!r}. "
            f"Must contain only letters, digits, and underscores."
        )
        raise ValueError(msg)


class Storage(Protocol, metaclass=ABCMeta):
    @abstractmethod
    async def startup(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def shutdown(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def get_schedules(self) -> Sequence[ScheduledJob]:
        raise NotImplementedError

    @abstractmethod
    async def add_schedule(self, *scheduled: ScheduledJob) -> None:
        raise NotImplementedError

    @abstractmethod
    async def delete_schedule(self, job_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    async def delete_schedule_many(self, job_ids: Sequence[str]) -> None:
        raise NotImplementedError



"""Module with all the main stuff for organizing and managing tasks in Jobify.

Includes JobRouter for grouping tasks into logical units, and RootRoute
and NodeRoute classes as basis for task execution and hierarchical routing.
"""

from jobify._internal.router.base import Route
from jobify._internal.router.node import NodeRoute
from jobify._internal.router.node import NodeRouter as JobRouter
from jobify._internal.router.root import RootRoute

__all__ = (
    "JobRouter",
    "NodeRoute",
    "RootRoute",
    "Route",
)



import asyncio
import sqlite3
from collections.abc import Callable, Sequence
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar
from zoneinfo import ZoneInfo

from typing_extensions import override

from jobify._internal.common.constants import EMPTY, JobStatus
from jobify._internal.storage.base import (
    ScheduledJob,
    Storage,
    validate_table_name,
)

if TYPE_CHECKING:
    from concurrent.futures import ThreadPoolExecutor

    from jobify._internal.common.types import LoopFactory


CREATE_SCHEDULED_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS {} (
    job_id TEXT PRIMARY KEY,
    name TEXT,
    message BLOB,
    status TEXT,
    next_run_at TEXT,
    created_at TEXT DEFAULT CURRENT_TIMESTAMP,
    updated_at TEXT DEFAULT CURRENT_TIMESTAMP
);
"""

SELECT_SCHEDULES_QUERY = """
SELECT job_id, name, message, status, next_run_at
FROM {};
"""

INSERT_SCHEDULE_QUERY = """
INSERT INTO {} (job_id, name, message, status, next_run_at)
VALUES (?, ?, ?, ?, ?)
ON CONFLICT (job_id) DO UPDATE SET
    name = EXCLUDED.name,
    message = EXCLUDED.message,
    status = EXCLUDED.status,
    next_run_at = EXCLUDED.next_run_at,
    updated_at = CURRENT_TIMESTAMP;
"""

DELETE_SCHEDULE_QUERY = """
DELETE FROM {} WHERE job_id = ?;
"""

DELETE_SCHEDULE_MANY_QUERY = """
DELETE FROM {table_name} WHERE job_id IN ({placeholder});
"""

ReturnT = TypeVar("ReturnT")

_Callback = Callable[[sqlite3.Connection], ReturnT]
_AsyncQueue: TypeAlias = asyncio.Queue[
    tuple[_Callback[Any], asyncio.Future[Any]],
]
_STOP: Any = object()


class SQLiteStorage(Storage):
    def __init__(
        self,
        database: str | Path = "jobify.db",
        *,
        table_name: str = "jobify_schedules",
        timeout: float = 20.0,
        max_queue_size: int = 1024,
    ) -> None:
        validate_table_name(table_name)
        self.database: Path = (
            Path(database) if isinstance(database, str) else database
        )
        self.table_name: str = table_name
        self.timeout: float = timeout
        self.tz: ZoneInfo = ZoneInfo("UTC")
        self.getloop: LoopFactory = asyncio._get_running_loop
        self.threadpool: ThreadPoolExecutor | None = None
        self.max_queue_size: int = max_queue_size

        self._queue: _AsyncQueue = EMPTY
        self._worker_task: asyncio.Task[None] | None = None

        self.create_scheduled_table_query: str = (
            CREATE_SCHEDULED_TABLE_QUERY.format(table_name)
        )
        self.select_schedules_query: str = SELECT_SCHEDULES_QUERY.format(
            table_name,
        )
        self.insert_schedule_query: str = INSERT_SCHEDULE_QUERY.format(
            table_name,
        )
        self.delete_schedule_query: str = DELETE_SCHEDULE_QUERY.format(
            table_name,
        )

    async def _worker(self, conn: sqlite3.Connection) -> None:
        loop = self.getloop()
        while True:
            item = await self._queue.get()

            if item is _STOP:
                self._queue.task_done()
                break

            callback, future = item
            try:
                result = await loop.run_in_executor(
                    self.threadpool,
                    callback,
                    conn,
                )
            except Exception as exc:  # noqa: BLE001
                future.set_exception(exc)
            else:
                future.set_result(result)
            finally:
                self._queue.task_done()

        conn.close()

    async def _execute(self, callback: _Callback[ReturnT]) -> ReturnT:
        loop = self.getloop()
        future: asyncio.Future[ReturnT] = loop.create_future()
        await self._queue.put((callback, future))
        return await future

    @override
    async def startup(self) -> None:
        conn = sqlite3.connect(
            database=self.database,
            timeout=self.timeout,
            check_same_thread=False,
        )
        _ = conn.execute("PRAGMA journal_mode=WAL;")
        _ = conn.execute("PRAGMA synchronous=NORMAL;")
        _ = conn.execute(self.create_scheduled_table_query)
        conn.commit()
        self._queue = asyncio.Queue(self.max_queue_size)
        self._worker_task = asyncio.create_task(self._worker(conn))

    @override
    async def shutdown(self) -> None:
        if self._queue is not EMPTY:
            await self._queue.put(_STOP)
            await self._queue.join()
            self._queue = EMPTY

        if self._worker_task is not None:
            await self._worker_task
            self._worker_task = None

    @override
    async def get_schedules(self) -> list[ScheduledJob]:
        def get(conn: sqlite3.Connection) -> list[ScheduledJob]:
            cursor = conn.execute(self.select_schedules_query)
            return [
                ScheduledJob(
                    job_id=row[0],
                    name=row[1],
                    message=row[2],
                    status=JobStatus(row[3]),
                    next_run_at=(
                        datetime.fromisoformat(row[4]).astimezone(self.tz)
                    ),
                )
                for row in cursor.fetchall()
            ]

        return await self._execute(get)

    @override
    async def add_schedule(self, *scheduled: ScheduledJob) -> None:
        def insert_many(conn: sqlite3.Connection) -> None:
            _ = conn.executemany(
                self.insert_schedule_query,
                [
                    (
                        sch.job_id,
                        sch.name,
                        sch.message,
                        sch.status,
                        sch.next_run_at.isoformat(),
                    )
                    for sch in scheduled
                ],
            )
            conn.commit()

        return await self._execute(insert_many)

    @override
    async def delete_schedule(self, job_id: str) -> None:
        def delete(conn: sqlite3.Connection) -> None:
            _ = conn.execute(self.delete_schedule_query, (job_id,))
            conn.commit()

        return await self._execute(delete)

    @override
    async def delete_schedule_many(self, job_ids: Sequence[str]) -> None:
        if not job_ids:
            return None

        job_ids = sorted(job_ids)

        def delete_many(conn: sqlite3.Connection) -> None:
            batch_size = 500
            for i in range(0, len(job_ids), batch_size):
                batch = job_ids[i : i + batch_size]
                query = DELETE_SCHEDULE_MANY_QUERY.format_map(
                    {
                        "table_name": self.table_name,
                        "placeholder": ",".join("?" * len(batch)),
                    }
                )
                _ = conn.execute(query, batch)
            conn.commit()

        return await self._execute(delete_many)



from __future__ import annotations

from typing import Any, Protocol, TypeVar

T = TypeVar("T")


class Loader(Protocol):
    def load(self, data: Any, tp: type[T], /) -> T:  # noqa: ANN401
        raise NotImplementedError


class Dumper(Protocol):
    def dump(self, data: Any, tp: Any, /) -> Any:  # noqa: ANN401
        raise NotImplementedError



# ruff: noqa: ANN401
from __future__ import annotations

from collections import UserDict
from typing import Any, Literal

from typing_extensions import Self, override


class EmptyPlaceholder(str):
    __slots__: tuple[()] = ()

    def __new__(cls) -> Self:
        return super().__new__(cls, "__EMPTY__")

    def __bool__(self) -> Literal[False]:
        return False

    @override
    def __str__(self) -> str:
        return super().__str__()

    @override
    def __hash__(self) -> int:
        return hash(super().__str__())

    @override
    def __eq__(self, other: object) -> bool:
        return other == super().__str__()


class State(UserDict[str, Any]):
    """An object that can be used to store arbitrary state."""

    data: dict[str, Any]
    __slots__: tuple[str, ...] = ("data",)

    def __init__(self, state: dict[str, Any] | None = None) -> None:  # pyright: ignore[reportMissingSuperCall]
        object.__setattr__(self, "data", state or {})

    @override
    def __setattr__(self, key: str, value: Any) -> None:
        self[key] = value

    def __getattr__(self, key: str) -> Any:
        try:
            return self.data[key]
        except KeyError as exc:
            message = (
                f"{self.__class__.__name__!r} object has no attribute {key!r}"
            )
            raise AttributeError(message) from exc

    @override
    def __delattr__(self, key: str) -> None:
        del self[key]

    @override
    def __str__(self) -> str:
        cls_name = type(self).__name__
        return f"{cls_name}({super().__str__()})"


class RequestState(State): ...



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, cast

from typing_extensions import override

from jobify._internal.router.base import Registrator, Route, Router

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence

    from jobify._internal.common.datastructures import State
    from jobify._internal.common.types import (
        ExceptionHandlers,
        Lifespan,
        MappingExceptionHandlers,
    )
    from jobify._internal.configuration import RouteOptions
    from jobify._internal.middleware.base import (
        BaseMiddleware,
        BaseOuterMiddleware,
    )
    from jobify._internal.scheduler.job import Job
    from jobify._internal.scheduler.scheduler import ScheduleBuilder


ReturnT = TypeVar("ReturnT")
ParamsT = ParamSpec("ParamsT")
NodeRouter_co = TypeVar("NodeRouter_co", bound="NodeRouter", covariant=True)


class NodeRoute(Route[ParamsT, ReturnT]):
    def __init__(
        self,
        *,
        name: str,
        func: Callable[ParamsT, ReturnT],
        options: RouteOptions,
    ) -> None:
        super().__init__(name, func, options)
        self._real_route: Route[ParamsT, ReturnT] | None = None

    @property
    def real_route(self) -> Route[ParamsT, ReturnT]:
        if self._real_route is None:
            msg = (
                f"Job {self.name!r} is not attached to any Jobify app."
                " Did you forget to call app.include_router()?"
            )
            raise RuntimeError(msg)
        return self._real_route

    def bind(self, route: Route[ParamsT, ReturnT]) -> None:
        self._real_route = route

    @override
    def schedule(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ScheduleBuilder[ReturnT]:
        return self.real_route.schedule(*args, **kwargs)

    @override
    async def push(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> Job[ReturnT]:
        return await self.real_route.push(*args, **kwargs)


class NodeRegistrator(Registrator[NodeRoute[..., Any]]):
    def __init__(  # noqa: PLR0913
        self,
        *,
        state: State | None,
        lifespan: Lifespan[NodeRouter_co] | None,
        route_class: type[NodeRoute[..., Any]],
        middleware: Sequence[BaseMiddleware] | None,
        outer_middleware: Sequence[BaseOuterMiddleware] | None,
        exception_handlers: MappingExceptionHandlers | None,
    ) -> None:
        super().__init__(
            state=state,
            lifespan=lifespan,
            route_class=route_class,
            middleware=middleware,
            outer_middleware=outer_middleware,
            exception_handlers=exception_handlers,
        )

    @override
    def register(
        self,
        name: str,
        func: Callable[ParamsT, ReturnT],
        options: RouteOptions,
    ) -> NodeRoute[ParamsT, ReturnT]:
        route = self.route_class(name=name, func=func, options=options)
        _ = functools.update_wrapper(route, func)
        self._routes[name] = route
        return cast("NodeRoute[ParamsT, ReturnT]", route)


class NodeRouter(Router):
    def __init__(  # noqa: PLR0913
        self,
        *,
        state: State | None = None,
        prefix: str | None = None,
        lifespan: Lifespan[NodeRouter_co] | None = None,
        middleware: Sequence[BaseMiddleware] | None = None,
        outer_middleware: Sequence[BaseOuterMiddleware] | None = None,
        exception_handlers: MappingExceptionHandlers | None = None,
        route_class: type[NodeRoute[..., Any]] = NodeRoute,
    ) -> None:
        super().__init__(prefix=prefix)
        self._registrator: NodeRegistrator = NodeRegistrator(
            state=state,
            lifespan=lifespan,
            middleware=middleware,
            outer_middleware=outer_middleware,
            exception_handlers=exception_handlers,
            route_class=route_class,
        )
        self.state: State = self._registrator.state
        self.exception_handlers: ExceptionHandlers = (
            self._registrator._exception_handlers
        )

    @property
    @override
    def task(self) -> NodeRegistrator:
        return self._registrator

    @property
    @override
    def routes(self) -> Iterator[NodeRoute[..., Any]]:
        yield from self.task._routes.values()

    @property
    @override
    def sub_routers(self) -> list[NodeRouter]:
        return cast("list[NodeRouter]", self._sub_routers)



"""Jobify entrypoint."""

from __future__ import annotations

import asyncio
import functools
import logging
import signal
import sys
import threading
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Literal, ParamSpec, TypeVar
from zoneinfo import ZoneInfo

from typing_extensions import Self

from jobify._internal.common.constants import EMPTY, PATCH_CRON_DEF_ID
from jobify._internal.configuration import (
    Cron,
    JobifyConfiguration,
    WorkerPools,
)
from jobify._internal.message import (
    AtArguments,
    CronArguments,
    Message,
    PushArguments,
    Triggers,
)
from jobify._internal.router.root import (
    CRONS_DEF_KEY,
    CronsDefinition,
    RootRoute,
    RootRouter,
)
from jobify._internal.scheduler.misfire_policy import (
    GracePolicy,
    MisfirePolicy,
    handle_misfire_policy,
)
from jobify._internal.serializers.json import JSONSerializer
from jobify._internal.serializers.json_extended import ExtendedJSONSerializer
from jobify._internal.storage.dummy import DummyStorage
from jobify._internal.storage.sqlite import SQLiteStorage
from jobify._internal.task_tracker import TaskTracker
from jobify._internal.typeadapter.dummy import DummyDumper, DummyLoader
from jobify.crontab import create_crontab

if TYPE_CHECKING:
    import inspect
    from collections.abc import Callable, Iterator, Sequence
    from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
    from datetime import datetime
    from types import FrameType, TracebackType

    from jobify._internal.common.datastructures import State
    from jobify._internal.common.types import (
        Lifespan,
        LoopFactory,
        MappingExceptionHandlers,
    )
    from jobify._internal.cron_parser import CronFactory
    from jobify._internal.middleware.base import (
        BaseMiddleware,
        BaseOuterMiddleware,
    )
    from jobify._internal.scheduler.job import Job
    from jobify._internal.scheduler.scheduler import ScheduleBuilder
    from jobify._internal.serializers.base import Serializer
    from jobify._internal.storage.base import ScheduledJob, Storage
    from jobify._internal.typeadapter.base import Dumper, Loader

HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)
if sys.platform == "win32":  # pragma: no cover
    # Windows signal 21. Sent by Ctrl+Break.
    HANDLED_SIGNALS += (signal.SIGBREAK,)  # pyright: ignore[reportConstantRedefinition]

logger = logging.getLogger("Jobify")

AppT = TypeVar("AppT", bound="Jobify")
ReturnT = TypeVar("ReturnT")
ParamsT = ParamSpec("ParamsT")


def cache_result(f: Callable[ParamsT, ReturnT]) -> Callable[ParamsT, ReturnT]:
    """Cache the result of the first function call."""
    result: ReturnT | None = None

    @functools.wraps(f)
    def wrapper(*args: ParamsT.args, **kwargs: ParamsT.kwargs) -> ReturnT:
        nonlocal result
        if result is None:
            result = f(*args, **kwargs)
        return result

    return wrapper


class Jobify(RootRouter):
    """Jobify is the main app for scheduling and managing background jobs.

    It provides a flexible and extensible framework for defining, running,
    and persisting jobs, supporting various executors, middleware, and
    serialization options.
    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        tz: ZoneInfo | None = None,
        state: State | None = None,
        dumper: Dumper | None = None,
        loader: Loader | None = None,
        storage: Storage | Literal[False] = EMPTY,
        lifespan: Lifespan[AppT] | None = None,
        serializer: Serializer | None = None,
        middleware: Sequence[BaseMiddleware] | None = None,
        outer_middleware: Sequence[BaseOuterMiddleware] | None = None,
        cron_factory: CronFactory = create_crontab,
        loop_factory: LoopFactory = asyncio.get_running_loop,
        exception_handlers: MappingExceptionHandlers | None = None,
        threadpool_executor: ThreadPoolExecutor | None = None,
        processpool_executor: ProcessPoolExecutor | None = None,
        route_class: type[RootRoute[..., Any]] = RootRoute,
    ) -> None:
        """Initialize a `Jobify` instance."""
        getloop = cache_result(loop_factory)
        tz = tz or ZoneInfo("UTC")

        if storage is False:
            storage = DummyStorage()
        elif storage is EMPTY:
            storage = SQLiteStorage()

        if isinstance(storage, SQLiteStorage):
            storage.getloop = getloop
            storage.threadpool = threadpool_executor
            storage.tz = tz

        system_types = (
            Message,
            Cron,
            PushArguments,
            AtArguments,
            CronArguments,
            MisfirePolicy,
            GracePolicy,
        )
        if serializer is None:
            serializer = (
                ExtendedJSONSerializer(system_types)
                if dumper is None and loader is None
                else JSONSerializer()
            )
        elif isinstance(serializer, ExtendedJSONSerializer):
            serializer.add_system_types(system_types)

        if dumper is None:
            dumper = DummyDumper()
        if loader is None:
            loader = DummyLoader()

        self.configs: JobifyConfiguration = JobifyConfiguration(
            tz=tz,
            dumper=dumper,
            loader=loader,
            storage=storage,
            getloop=getloop,
            serializer=serializer,
            worker_pools=WorkerPools(
                _processpool=processpool_executor,
                threadpool=threadpool_executor,
            ),
            cron_factory=cron_factory,
        )
        idle_event = asyncio.Event()
        idle_event.set()
        super().__init__(
            state=state,
            lifespan=lifespan,
            middleware=middleware,
            outer_middleware=outer_middleware,
            task_tracker=TaskTracker(
                pending_jobs={},
                pending_tasks={},
                idle_event=idle_event,
            ),
            jobify_config=self.configs,
            exception_handlers=exception_handlers,
            route_class=route_class,
        )
        self._captured_signals: list[int] = []

    def find_job(self, id_: str, /) -> Job[ReturnT] | None:
        """Find an active job by its ID.

        Args:
            id_: Unique identifier of the job.

        Returns:
            The `Job` instance if it's currently pending or running,
            otherwise `None`.

        """
        return self.task._task_tracker.pending_jobs.get(id_)

    def get_active_jobs(self) -> list[Job[Any]]:
        """Return a list of all currently active jobs."""
        return list(self.task._task_tracker.pending_jobs.values())

    async def __aenter__(self) -> Self:
        """Enter the Jobify context manager.

        Returns:
            The initialized Jobify instance ready for use.

        Raises:
            Any exception raised by `startup()` method.

        """
        await self.startup()
        return self

    async def startup(self) -> None:
        """Initialize the Jobify application.

        This method:
        1. Marks the application as started
        2. Propagates startup events to all routers and their registrators
        3. Schedules any pending cron jobs

        Raises:
            RuntimeError: If application startup fails due to configuration
            issues or router initialization errors.

        """
        self.propagate_real_routes()
        self.configs.app_started = True

        await self.configs.storage.startup()
        await self._propagate_startup(self)
        await self._restore_schedules()

        logger.info("Jobify startup complete. Ready to schedule jobs.")

    async def _restore_schedules(self) -> None:
        """Restore schedules.

        Phase 1: Synchronize Declarative Cron Jobs (Code and Database)
            - If there are changes or new jobs, run the full process
              (middleware and database persistence).
            - If no changes, silently restore jobs in memory.
        Phase 2: Restore Imperative Jobs (Database Only)
            - Restore valid jobs and delete obsolete ones from the database.
        """
        crons_def: CronsDefinition = self.task.state.pop(CRONS_DEF_KEY, {})
        schedules = await self.configs.storage.get_schedules()

        to_delete: list[str] = []
        processed_jobs: set[str] = set()
        db_map = {sch.job_id: sch for sch in schedules}

        # --- PHASE 1: Declarative Crons ---
        for job_id, (rout, cron_def) in crons_def.items():
            processed_jobs.add(job_id)
            builder = rout.schedule(*cron_def.args, **cron_def.kwargs)

            if (sch_in_db := db_map.get(job_id)) is not None:
                restored = self._restore_job_from_storage(sch_in_db)
                if (
                    restored is not None
                    and (msg := restored[1])
                    and isinstance(msg.trigger, CronArguments)
                ):
                    self._start_restored_job_in_memory(
                        restored[0],
                        msg.trigger,
                        sch_in_db.next_run_at,
                    )
                    if msg.trigger.cron == cron_def:
                        continue

            _ = await builder.cron(
                cron_def,
                job_id=job_id,
                replace=True,
                force=True,
            )

        # --- PHASE 2: Dynamic/Imperative Jobs & Cleanup ---
        for job_id, sch in db_map.items():
            if job_id in processed_jobs:
                continue

            if job_id.endswith(PATCH_CRON_DEF_ID):
                to_delete.append(job_id)
                continue

            if (restored := self._restore_job_from_storage(sch)) is not None:
                builder, msg, _ = restored
                self._start_restored_job_in_memory(
                    builder,
                    msg.trigger,
                    sch.next_run_at,
                )
                continue

            to_delete.append(job_id)

        if to_delete:
            logger.info("Cleaning up %d obsolete/broken jobs", len(to_delete))
            await self.configs.storage.delete_schedule_many(to_delete)

    def _restore_job_from_storage(
        self,
        sch: ScheduledJob,
    ) -> tuple[ScheduleBuilder[Any], Message, inspect.BoundArguments] | None:
        """Safely deserializes and binds a job from storage.

        Returns:
            Tuple (Builder, Message, BoundArguments) or None if failed.

        """
        route = self.task._routes.get(sch.name)
        if not route or route.options.get("durable") is False:
            return None
        try:
            raw_msg = self.configs.serializer.loadb(sch.message)
            msg = self.configs.loader.load(raw_msg, Message)

            for name, raw_arg in msg.arguments.items():
                param_type = route.func_spec.type_params[name]
                msg.arguments[name] = self.configs.loader.load(
                    raw_arg,
                    param_type,
                )
            bound = route.func_spec.signature.bind(**msg.arguments)
        except (KeyError, TypeError, ValueError) as exc:
            logger.warning(
                "Failed to revive job %s (%s). Reason: %s",
                sch.job_id,
                sch.name,
                exc,
            )
            return None
        else:
            builder = route.create_builder(bound)
            return builder, msg, bound

    def _start_restored_job_in_memory(
        self,
        builder: ScheduleBuilder[Any],
        trigger: Triggers,
        db_next_run_at: datetime,
    ) -> None:
        """Start the internal timer for a restored job.

        IMPORTANT: This bypasses middleware and database persistence.
        It handles misfire policy to ensure we do not execute old jobs
        incorrectly.
        """
        match trigger:
            case PushArguments():
                builder._push(trigger.job_id, exec_at=db_next_run_at)
            case AtArguments():
                builder._at(trigger.at, trigger.job_id)
            case CronArguments():
                parser = self.configs.cron_factory(trigger.cron.expression)
                next_run_at = handle_misfire_policy(
                    cron_parser=parser,
                    next_run_at=db_next_run_at,
                    real_now=builder.now(),
                    policy=trigger.cron.misfire_policy,
                )
                builder._cron(
                    cron=trigger.cron,
                    job_id=trigger.job_id,
                    offset=trigger.offset,
                    next_run_at=next_run_at,
                    cron_parser=parser,
                    run_count=trigger.run_count,
                )

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        """Exit the Jobify context manager.

        Note:
            This method ensures proper shutdown regardless of whether an
            exception occurred in the managed context. The exception parameters
            are ignored as shutdown should proceed even if the context failed.

        """
        await self.shutdown()

    async def shutdown(self) -> None:
        """Gracefully shut down the Jobify application.

        This method performs a structured shutdown:
        1. Marks the application as stopped (`app_started = False`).
        2. Propagates shutdown events to all routers/components.
        3. Cancels all scheduled future jobs in the registry
           (`_jobs_registry`).
        4. Closes the jobify configuration (e.g., stopping the internal
           scheduler).
        5. Cancels all currently running tasks (in `_tasks_registry`), waits
           for their completion, and explicitly clears the task registry.

        Note:
            The method uses `return_exceptions=True` when gathering cancelled
            tasks to prevent shutdown from being interrupted by task exception.

        """
        self.configs.app_started = False

        if jobs := tuple(self.task._task_tracker.pending_jobs.values()):
            for job in jobs:
                job._cancel()

        if (
            tasks := self.task._task_tracker.pending_tasks.values()
        ):  # pragma: no cover
            for task in tuple(tasks):
                _ = task.cancel()
            _ = await asyncio.gather(*tasks, return_exceptions=True)

        self.configs.worker_pools.close()
        await self._propagate_shutdown()
        await self.configs.storage.shutdown()

        # If we did gracefully shut down due to a signal, try to
        # trigger the expected behaviour now; multiple signals would be
        # done LIFO, see https://stackoverflow.com/questions/48434964
        for captured_signal in reversed(self._captured_signals):
            signal.raise_signal(captured_signal)

        logger.info("Jobify shutdown complete.")

    async def wait_all(self, timeout: float | None = None) -> None:
        """Wait for all currently scheduled jobs to complete.

        This method waits until all jobs currently registered have finished
        executing (with statuses of SUCCESS, FAILED, or TIMEOUT). This is
        useful in situations where it's important to ensure that background
        tasks have completed before moving on.

        The method sets an internal event when both conditions are met:
        1. No jobs remain in the jobs registry (`_jobs_registry`)

        Args:
            timeout (optional): The maximum time in seconds to wait for the
                jobs to complete. If not specified, the default value of `None`
                will be used, which means the job will wait indefinitely. If a
                timeout is specified and it is reached, the method will raise
                an `asyncio.TimeoutError`.

        """
        idle_event = self.task._task_tracker.idle_event
        with self._capture_signals():
            _ = await asyncio.wait_for(idle_event.wait(), timeout=timeout)

    @contextmanager
    def _capture_signals(self) -> Iterator[None]:
        # Signals can only be listened to from the main thread.
        if threading.current_thread() is not threading.main_thread():
            yield
            return
        # always use signal.signal, even if loop.add_signal_handler is
        # available this allows to restore previous signal handlers later on
        original_handlers = {
            sig: signal.signal(sig, self._handle_exit)
            for sig in HANDLED_SIGNALS
        }
        try:
            yield
        finally:
            for sig, handler in original_handlers.items():
                _ = signal.signal(sig, handler)

    def _handle_exit(self, sig: int, _: FrameType | None) -> None:
        self._captured_signals.append(sig)
        loop = self.configs.getloop()
        idle_event = self.task._task_tracker.idle_event
        _handle = loop.call_soon_threadsafe(idle_event.set)



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Generic, TypeVar, final

from typing_extensions import override

from jobify._internal.common.constants import EMPTY, INFINITY, JobStatus
from jobify._internal.exceptions import JobFailedError, JobNotCompletedError

if TYPE_CHECKING:
    from collections.abc import Callable, Generator
    from datetime import datetime

    from jobify._internal.configuration import Cron
    from jobify._internal.cron_parser import CronParser
    from jobify._internal.storage.base import Storage

ReturnT = TypeVar("ReturnT")


@final
class CronContext(Generic[ReturnT]):
    __slots__: tuple[str, ...] = (
        "cron",
        "cron_parser",
        "failure_count",
        "job",
        "offset",
        "run_count",
    )

    def __init__(  # noqa: PLR0913
        self,
        *,
        cron: Cron,
        cron_parser: CronParser,
        failure_count: int = 0,
        job: Job[ReturnT],
        offset: datetime,
        run_count: int,
    ) -> None:
        self.cron = cron
        self.cron_parser = cron_parser
        self.failure_count = failure_count
        self.job = job
        self.offset = offset
        self.run_count = run_count

    def is_run_exceeded_by_limit(self) -> bool:
        if self.cron.max_runs == INFINITY:
            return False
        return self.run_count >= self.cron.max_runs

    def is_failure_allowed_by_limit(self) -> bool:
        return self.failure_count < self.cron.max_failures


@final
class Job(Generic[ReturnT]):
    __slots__: tuple[str, ...] = (
        "_cron_context",
        "_event",
        "_handle",
        "_result",
        "_storage",
        "_unregister_hook",
        "exception",
        "exec_at",
        "id",
        "status",
    )

    def __init__(
        self,
        *,
        job_id: str,
        storage: Storage,
        exec_at: datetime,
        unregister_hook: Callable[[str], None],
        job_status: JobStatus = JobStatus.PENDING,
    ) -> None:
        self._unregister_hook = unregister_hook
        self._event = asyncio.Event()
        self._result: ReturnT = EMPTY
        self._storage = storage
        self._handle: asyncio.Handle | None = None
        self._cron_context: CronContext[ReturnT] | None = None
        self.id = job_id
        self.status = job_status
        self.exception: Exception | None = None
        self.exec_at = exec_at

        self._event.set()

    @property
    def cron_expression(self) -> str | None:
        if self._cron_context is not None:
            return self._cron_context.cron.expression
        return None

    @override
    def __repr__(self) -> str:
        if self.cron_expression is not None:
            cron_info = f", cron={self.cron_expression!r}"
        else:
            cron_info = ""
        return (
            f"<{type(self).__name__} "
            f"id={self.id!r}, "
            f"status={self.status.value!r}, "
            f"exec_at={self.exec_at.isoformat()!r}"
            f"{cron_info}>"
        )

    def __await__(self) -> Generator[object, None, ReturnT]:
        async def _await() -> ReturnT:
            await self.wait()
            return self.result()

        return _await().__await__()

    def bind_handle(self, handle: asyncio.Handle) -> None:
        self._handle = handle

    def bind_cron_context(self, ctx: CronContext[ReturnT]) -> None:
        self._cron_context = ctx

    def result(self) -> ReturnT:
        if self.status is JobStatus.SUCCESS or self._result is not EMPTY:
            return self._result
        if self.status is JobStatus.FAILED:
            raise JobFailedError(
                self.id,
                reason=str(self.exception),
            ) from self.exception
        raise JobNotCompletedError

    def set_result(self, val: ReturnT, *, status: JobStatus) -> None:
        self._result = val
        self.status = status

    def set_exception(self, exc: Exception, *, status: JobStatus) -> None:
        self.exception = exc
        self.status = status

    def update(self, *, exec_at: datetime, status: JobStatus) -> None:
        self._event = asyncio.Event()
        self.status = status
        self.exec_at = exec_at

    def is_done(self) -> bool:
        return self._event.is_set()

    def is_cron(self) -> bool:
        return self._cron_context is not None

    def is_reschedulable(self) -> bool:
        return self.status not in (
            JobStatus.PERMANENTLY_FAILED,
            JobStatus.CANCELLED,
        )

    async def wait(self) -> None:
        """Wait until the job is done.

        If the job is already completed, this method returns immediately.
        Safe for concurrent use by multiple coroutines.
        """
        _ = await self._event.wait()

    async def cancel(self) -> None:
        self.status = JobStatus.CANCELLED
        self._cancel()
        await self._storage.delete_schedule(self.id)

    def _cancel(self) -> None:
        self._event.set()
        self._unregister_hook(self.id)
        if self._handle is not None:
            self._handle.cancel()



from typing import NoReturn


class BaseJobifyError(Exception):
    pass


class JobNotCompletedError(BaseJobifyError):
    """Raised when trying to access result of incomplete job."""

    def __init__(
        self,
        msg: str = (
            "Job result is not ready yet, "
            "please use .wait() and then you can use .result"
        ),
    ) -> None:
        super().__init__(msg)


class JobFailedError(BaseJobifyError):
    def __init__(self, job_id: str, reason: str) -> None:
        self.job_id: str = job_id
        self.reason: str = reason
        super().__init__(f"job_id: {job_id}, failed_reason: {reason}")


class JobTimeoutError(BaseJobifyError):
    """Raised when job execution exceeds the configured timeout."""

    def __init__(self, job_id: str, timeout: float) -> None:
        self.job_id: str = job_id
        self.timeout: float = timeout

        msg = (
            f"job_id: {job_id} exceeded timeout of {timeout} seconds. "
            "Job execution was interrupted."
        )
        super().__init__(msg)


class DuplicateJobError(RuntimeError):
    """Raised when a job is scheduled with an ID that is already in use."""

    def __init__(self, job_id: str) -> None:
        self.job_id: str = job_id
        msg = (
            f"Job with ID {job_id!r} is already scheduled. "
            "Use 'replace=True' if you want to update the existing job."
        )
        super().__init__(msg)


class RouteAlreadyRegisteredError(BaseJobifyError):
    """A route with this name has already been registered."""

    def __init__(self, name: str) -> None:
        msg = f"A route with the name {name!r} has already been registered."
        super().__init__(msg)


class ApplicationStateError(BaseJobifyError):
    """Raised when app is in wrong state for the requested operation."""

    def __init__(
        self,
        *,
        operation: str,
        reason: str,
        solution: str,
    ) -> None:
        self.operation: str = operation
        self.reason: str = reason
        self.solution: str = solution

        msg = (
            f"Cannot perform operation '{operation}'.\n"
            f"  Reason: {reason}\n"
            f"  Resolution: {solution}"
        )
        super().__init__(msg)


def raise_app_not_started_error(operation: str) -> NoReturn:
    raise ApplicationStateError(
        operation=operation,
        reason="The Jobify application is not started.",
        solution=(
            "Ensure you are calling this method inside an 'async with jobify:'"
            "block or after calling 'await jobify.startup()'."
        ),
    )


def raise_app_already_started_error(operation: str) -> NoReturn:
    raise ApplicationStateError(
        operation=operation,
        reason="The Jobify app's already running and configuration is frozen.",
        solution=(
            "Configuration methods (register, add_middleware, etc.)"
            "must be called BEFORE the application starts."
            "Move this call outside/before the 'async with jobify:' block."
        ),
    )


class NoResultError(BaseJobifyError):
    """Raised when a task should fail immediately without retrying."""

    def __init__(
        self, msg: str = "Job aborted: no result expected and retries stopped."
    ) -> None:
        super().__init__(msg)



"""Middleware system for job execution pipeline.

This module provides a basic interface for creating middleware that can
intercept and process jobs before they reach their final destination.
Middleware can be combined together to create a processing chain,
where each piece of middleware can:

- Execute code before the job is handled
- Pass control to the next middleware in the pipeline
- Execute code after the job has been handled
- Modify the state or the result
- Break out of the chain by skipping the `call_next()` method
"""

from jobify._internal.middleware.base import (
    BaseMiddleware,
    BaseOuterMiddleware,
    CallNext,
    CallNextOuter,
)

__all__ = (
    "BaseMiddleware",
    "BaseOuterMiddleware",
    "CallNext",
    "CallNextOuter",
)



from collections.abc import Sequence

from typing_extensions import override

from jobify._internal.storage.base import ScheduledJob, Storage


class DummyStorage(Storage):
    @override
    async def startup(self) -> None:
        pass

    @override
    async def shutdown(self) -> None:
        pass

    @override
    async def get_schedules(self) -> Sequence[ScheduledJob]:
        return []

    @override
    async def add_schedule(self, *scheduled: ScheduledJob) -> None:
        pass

    @override
    async def delete_schedule(self, job_id: str) -> None:
        pass

    @override
    async def delete_schedule_many(self, job_ids: Sequence[str]) -> None:
        pass



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, final

from typing_extensions import override

from jobify._internal.middleware.base import BaseMiddleware, CallNext

if TYPE_CHECKING:
    from jobify._internal.configuration import JobifyConfiguration
    from jobify._internal.context import JobContext


@final
class ExceptionMiddleware(BaseMiddleware):
    __slots__: tuple[str, ...] = ("jobify_config",)

    def __init__(self, jobify_config: JobifyConfiguration) -> None:
        self.jobify_config = jobify_config

    @override
    async def __call__(self, call_next: CallNext, context: JobContext) -> Any:
        try:
            return await call_next(context)
        except Exception as exc:
            for cls_exc in type(exc).__mro__:
                exc_handlers = context.schedule_builder._exception_handlers
                if handler := exc_handlers.get(cls_exc):
                    break
            else:
                raise

            if asyncio.iscoroutinefunction(handler):
                return await handler(exc, context)

            loop = self.jobify_config.getloop()
            thread = self.jobify_config.worker_pools.threadpool
            return await loop.run_in_executor(thread, handler, exc, context)



"""Core scheduling components for the jobify framework.

This module provides the basic classes for job scheduling and management.
It exposes the main scheduler interface and job planning components that
form the basis of the jobify asynchronous job scheduling system.
"""

from importlib.metadata import version as get_version

from jobify._internal.common.constants import JobStatus, RunMode
from jobify._internal.common.datastructures import RequestState, State
from jobify._internal.configuration import Cron
from jobify._internal.context import JobContext, OuterContext
from jobify._internal.inspection import INJECT
from jobify._internal.router.node import NodeRouter as JobRouter
from jobify._internal.runners import Runnable
from jobify._internal.scheduler.job import CronContext, Job
from jobify._internal.scheduler.misfire_policy import (
    GracePolicy,
    MisfirePolicy,
)
from jobify._internal.scheduler.scheduler import ScheduleBuilder
from jobify.jobify import Jobify

__version__ = get_version("jobify")
__all__ = (
    "INJECT",
    "Cron",
    "CronContext",
    "GracePolicy",
    "Job",
    "JobContext",
    "JobRouter",
    "JobStatus",
    "Jobify",
    "MisfirePolicy",
    "OuterContext",
    "RequestState",
    "RunMode",
    "Runnable",
    "ScheduleBuilder",
    "State",
)



from __future__ import annotations

import functools
import inspect
import warnings
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, Final, Generic, ParamSpec, TypeVar

from typing_extensions import override

from jobify._internal.common.constants import RunMode

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from concurrent.futures import Executor

    from jobify._internal.common.types import LoopFactory
    from jobify._internal.configuration import JobifyConfiguration
    from jobify._internal.inspection import FuncSpec

ReturnT = TypeVar("ReturnT")
ParamsT = ParamSpec("ParamsT")


class RunStrategy(ABC, Generic[ParamsT, ReturnT]):
    __slots__: tuple[str, ...] = ("func",)

    def __init__(self, func: Callable[ParamsT, ReturnT]) -> None:
        self.func: Final = func

    @abstractmethod
    async def __call__(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ReturnT:
        raise NotImplementedError


class SyncStrategy(RunStrategy[ParamsT, ReturnT]):
    @override
    async def __call__(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ReturnT:
        return self.func(*args, **kwargs)


class AsyncStrategy(RunStrategy[ParamsT, ReturnT]):
    @override
    async def __call__(
        self: AsyncStrategy[ParamsT, Awaitable[ReturnT]],
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ReturnT:
        return await self.func(*args, **kwargs)


class PoolStrategy(RunStrategy[ParamsT, ReturnT]):
    __slots__: tuple[str, ...] = ("executor", "getloop")

    def __init__(
        self,
        func: Callable[ParamsT, ReturnT],
        executor: Executor | None,
        getloop: LoopFactory,
    ) -> None:
        super().__init__(func)
        self.executor: Executor | None = executor
        self.getloop: LoopFactory = getloop

    @override
    async def __call__(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ReturnT:
        func_call = functools.partial(self.func, *args, **kwargs)
        return await self.getloop().run_in_executor(self.executor, func_call)


class Runnable(Generic[ReturnT]):
    __slots__: tuple[str, ...] = (
        "bound",
        "func_spec",
        "name",
        "origin_arguments",
        "strategy",
    )

    def __init__(
        self,
        *,
        name: str,
        bound: inspect.BoundArguments,
        strategy: RunStrategy[ParamsT, ReturnT],
        func_spec: FuncSpec[ReturnT],
    ) -> None:
        self.name: str = name
        self.strategy: Final = strategy
        self.bound: inspect.BoundArguments = bound
        self.origin_arguments: dict[str, Any] = bound.arguments.copy()
        self.func_spec: FuncSpec[ReturnT] = func_spec

    def __call__(self) -> Awaitable[ReturnT]:
        return self.strategy(*self.bound.args, **self.bound.kwargs)


def _validate_run_mode(mode: RunMode | None, *, is_async: bool) -> RunMode:
    if is_async:
        if mode in (RunMode.PROCESS, RunMode.THREAD):
            msg = (
                "Async functions are always done in the main loop."
                " This mode (PROCESS/THREAD) is not used."
            )
            warnings.warn(msg, category=RuntimeWarning, stacklevel=3)
        return RunMode.MAIN
    if mode is None:
        return RunMode.THREAD
    return mode


def create_run_strategy(
    func: Callable[ParamsT, ReturnT],
    jobify_config: JobifyConfiguration,
    *,
    mode: RunMode | None,
) -> RunStrategy[ParamsT, ReturnT]:
    # inspect.iscoroutinefunction returns TypeGuard,
    # but we need a regular bool variable
    is_async = bool(inspect.iscoroutinefunction(func))

    mode = _validate_run_mode(mode, is_async=is_async)
    if is_async:
        return AsyncStrategy(func)

    match mode:
        case RunMode.PROCESS:
            processpool = jobify_config.worker_pools.processpool
            return PoolStrategy(func, processpool, jobify_config.getloop)
        case RunMode.THREAD:
            threadpool = jobify_config.worker_pools.threadpool
            return PoolStrategy(func, threadpool, jobify_config.getloop)
        case _:
            return SyncStrategy(func)



from datetime import datetime, timedelta
from enum import Enum
from typing import NamedTuple

from typing_extensions import override

from jobify._internal.cron_parser import CronParser


class GracePolicy(NamedTuple):
    value: timedelta

    @override
    def __str__(self) -> str:
        return f"<{type(self).__name__} value={self.value}>"


class MisfirePolicy(str, Enum):
    ALL = "all"
    SKIP = "skip"
    ONCE = "once"

    @staticmethod
    def GRACE(t: timedelta, /) -> GracePolicy:  # noqa: N802
        return GracePolicy(t)


def handle_misfire_policy(
    *,
    cron_parser: CronParser,
    next_run_at: datetime,
    real_now: datetime,
    policy: MisfirePolicy | GracePolicy,
) -> datetime:
    if next_run_at >= real_now:
        return next_run_at

    match policy:
        case MisfirePolicy.ALL:
            return next_run_at
        case MisfirePolicy.SKIP:
            return cron_parser.next_run(now=real_now)
        case MisfirePolicy.ONCE:
            return real_now
        case GracePolicy(value=grace_delta):
            cutoff = real_now - grace_delta
            if next_run_at >= cutoff:
                return next_run_at
            return cron_parser.next_run(now=cutoff)



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, NamedTuple

if TYPE_CHECKING:
    import asyncio

    from jobify._internal.scheduler.job import Job


class TaskTracker(NamedTuple):
    pending_jobs: dict[str, Job[Any]]
    pending_tasks: dict[str, asyncio.Task[Any]]
    idle_event: asyncio.Event

    def register_job(self, job: Job[Any]) -> None:
        job._event.clear()
        self.idle_event.clear()
        self.pending_jobs[job.id] = job

    def check_state(self) -> None:
        if not (self.pending_jobs or self.pending_tasks):
            self.idle_event.set()

    def unregister_job(self, job_id: str) -> None:
        if task := self.pending_tasks.pop(job_id, None):
            _ = task.cancel()
        if self.pending_jobs.pop(job_id, None):
            self.check_state()

    def track_task(
        self,
        job_id: str,
        task: asyncio.Task[Any],
        event: asyncio.Event,
    ) -> None:
        self.idle_event.clear()
        self.pending_tasks[job_id] = task
        part = functools.partial(
            self._on_task_done,
            job_id=job_id,
            event=event,
        )
        task.add_done_callback(part)

    def _on_task_done(
        self,
        _t: asyncio.Task[Any],
        *,
        job_id: str,
        event: asyncio.Event,
    ) -> None:
        _ = self.pending_tasks.pop(job_id, None)
        self.check_state()
        event.set()



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any

from typing_extensions import override

from jobify._internal.exceptions import NoResultError
from jobify._internal.middleware.base import BaseMiddleware, CallNext

if TYPE_CHECKING:
    from jobify._internal.context import JobContext

logger = logging.getLogger("jobify.middleware")


class RetryMiddleware(BaseMiddleware):
    @override
    async def __call__(self, call_next: CallNext, context: JobContext) -> Any:
        max_retries = context.route_options.get("retry")
        if max_retries is None:
            return await call_next(context)

        failures = 0
        while True:
            try:
                return await call_next(context)
            except NoResultError:  # noqa: PERF203
                raise
            except Exception as exc:
                failures += 1
                if failures > max_retries:
                    msg = (
                        f"Job failed after exhausting all {max_retries}"
                        " retries. Propagating error."
                    )
                    logger.warning(msg)
                    raise

                seconds_wait = min(2 ** (failures - 1), 60)
                logger.warning(
                    "Attempt %s/%s failed. Retrying in %ss. Error: %s",
                    failures,
                    max_retries,
                    seconds_wait,
                    exc,
                )
                await asyncio.sleep(seconds_wait)  # Exponential backoff



"""Serializers module for Jobify.

This module provides serialization utilities for task data and results.
Available serializers implement the JobsSerializer protocol.

Classes:
    obsSerializer: Abstract base protocol defining serializer interface
    AstLiteralSerializer: Safe serializer using AST literal evaluation
    UnsafePickleSerializer: Pickle-based serializer (use with caution)

Protocol Interface:
    dumpb(value: Any) -> bytes: Serialize object to bytes
    loadb(value: bytes) -> Any: Deserialize bytes to object

Security Notes:
    - AstLiteralSerializer: Safe for untrusted data, but limited to basic Python literals
    - UnsafePickleSerializer: UNSAFE for untrusted data - allows arbitrary code execution

"""  # noqa: E501

from jobify._internal.serializers.base import Serializer
from jobify._internal.serializers.json import JSONSerializer
from jobify._internal.serializers.json_extended import ExtendedJSONSerializer
from jobify._internal.serializers.pickle_unsafe import UnsafePickleSerializer

__all__ = (
    "ExtendedJSONSerializer",
    "JSONSerializer",
    "Serializer",
    "UnsafePickleSerializer",
)



"""Type adaptation and serialization utilities.

This module provides base classes and interfaces for converting Python types
to serializable formats and back. It serves as the foundation for type
adaptation, leveraging implementations from pydantic and adaptix.

The module exports the main type adaptation interfaces:
- Dumper: Base class for serializing Python objects to JSON-compatible types
- Loader: Base class for deserializing JSON-compatible types to Python objects
"""

from jobify._internal.typeadapter.base import Dumper, Loader

__all__ = ("Dumper", "Loader")



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any, Protocol, TypeAlias

JSONCompat: TypeAlias = (
    dict[str, "JSONCompat"]
    | list["JSONCompat"]
    | str
    | int
    | float
    | bool
    | None
)


class Serializer(Protocol, metaclass=ABCMeta):
    @abstractmethod
    def dumpb(self, data: Any) -> bytes:  # noqa: ANN401
        raise NotImplementedError

    @abstractmethod
    def loadb(self, data: bytes) -> Any:  # noqa: ANN401
        raise NotImplementedError



"""Custom exceptions for the jobify framework.

This module defines specific exceptions that the jobify scheduling
system can raise. These exceptions provide more detailed information
about errors and guidance on how to handle common scheduling scenarios.
"""

from jobify._internal.exceptions import (
    ApplicationStateError,
    BaseJobifyError,
    DuplicateJobError,
    JobFailedError,
    JobNotCompletedError,
    JobTimeoutError,
    NoResultError,
    RouteAlreadyRegisteredError,
)

__all__ = (
    "ApplicationStateError",
    "BaseJobifyError",
    "DuplicateJobError",
    "JobFailedError",
    "JobNotCompletedError",
    "JobTimeoutError",
    "NoResultError",
    "RouteAlreadyRegisteredError",
)



import inspect
from collections.abc import Callable
from typing import (
    Any,
    Generic,
    ParamSpec,
    TypeAlias,
    TypeVar,
    final,
    get_origin,
    get_type_hints,
)

ReturnT = TypeVar("ReturnT")
ParamsT = ParamSpec("ParamsT")

INJECT: Any = object()
ParamName: TypeAlias = str
TypeHint: TypeAlias = Any


@final
class FuncSpec(Generic[ReturnT]):
    __slots__: tuple[str, ...] = (
        "inject_params",
        "name",
        "result_type",
        "signature",
        "type_params",
    )

    def __init__(
        self,
        *,
        name: str,
        signature: inspect.Signature,
        result_type: type[ReturnT],
        type_params: dict[ParamName, TypeHint],
        inject_params: dict[str, Any],
    ) -> None:
        self.name = name
        self.signature = signature
        self.result_type = result_type
        self.type_params = type_params
        self.inject_params = inject_params


def get_type_params(
    sig: inspect.Signature,
    hints: dict[str, Any],
) -> dict[ParamName, TypeHint]:
    return {
        arg.name: hints.get(arg.name, Any) for arg in sig.parameters.values()
    }


def get_inject_params(
    sig: inspect.Signature,
    hints: dict[str, Any],
) -> dict[str, Any]:
    inject_params: dict[str, Any] = {}
    for name, param in sig.parameters.items():
        if param.default is not INJECT:
            continue
        annotation = hints.get(name, inspect.Parameter.empty)
        if annotation is inspect.Parameter.empty:
            msg = f"Parameter {name!r} requires a type annotation for INJECT"
            raise ValueError(msg)
        inject_params[name] = get_origin(annotation) or annotation

    return inject_params


def get_result_type(hints: dict[str, Any]) -> Any:  # noqa: ANN401
    return hints.get("return", Any)


def make_func_spec(func: Callable[ParamsT, ReturnT]) -> FuncSpec[ReturnT]:
    sig = inspect.signature(func)
    hints = get_type_hints(func)
    return FuncSpec(
        name=func.__name__,
        result_type=get_result_type(hints),
        type_params=get_type_params(sig, hints),
        inject_params=get_inject_params(sig, hints),
        signature=sig,
    )



"""Cron Parser implementation."""

from datetime import datetime
from typing import Final

from crontab import CronTab as _CronTab
from typing_extensions import override

from jobify._internal.cron_parser import CronParser


class CronTab(CronParser):
    """Cron expression parser based on the `crontab` library."""

    __slots__: tuple[str, ...] = ("_entry",)

    def __init__(self, expression: str) -> None:
        """Initialize a CronTab parser."""
        self._entry: Final = _CronTab(expression)

    @override
    def next_run(self, *, now: datetime) -> datetime:
        """Compute the next scheduled execution time."""
        return self._entry.next(now=now, return_datetime=True)  # type: ignore[no-any-return] # pyright: ignore[reportAttributeAccessIssue,reportUnknownMemberType,reportUnknownVariableType]


def create_crontab(expression: str) -> CronTab:
    """Create a CronTab instance."""
    return CronTab(expression)



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any

from typing_extensions import override

from jobify._internal.exceptions import JobTimeoutError
from jobify._internal.middleware.base import BaseMiddleware, CallNext

if TYPE_CHECKING:
    from jobify._internal.context import JobContext

logger = logging.getLogger("jobify.middleware")


class TimeoutMiddleware(BaseMiddleware):
    @override
    async def __call__(self, call_next: CallNext, context: JobContext) -> Any:
        timeout = context.route_options.get("timeout")
        if timeout is None:
            return await call_next(context)

        try:
            return await asyncio.wait_for(call_next(context), timeout=timeout)
        except asyncio.TimeoutError as exc:
            job_id = context.job.id
            raise JobTimeoutError(job_id=job_id, timeout=timeout) from exc



from enum import Enum, unique
from typing import Any

from jobify._internal.common.datastructures import EmptyPlaceholder

EMPTY: Any = EmptyPlaceholder()
INFINITY = -1
PATCH_FUNC_NAME = "__jobify_original"
PATCH_CRON_DEF_ID = "__jobify_cron_definition"


@unique
class JobStatus(str, Enum):
    PENDING = "pending"
    SCHEDULED = "scheduled"
    RUNNING = "running"
    CANCELLED = "cancelled"
    SUCCESS = "success"
    FAILED = "failed"
    TIMEOUT = "timeout"
    PERMANENTLY_FAILED = "permanently_failed"


@unique
class RunMode(str, Enum):
    MAIN = "main"
    THREAD = "thread"
    PROCESS = "process"



import json

from typing_extensions import override

from jobify._internal.serializers.base import JSONCompat, Serializer


class JSONSerializer(Serializer):
    @override
    def dumpb(self, data: JSONCompat) -> bytes:
        return json.dumps(data).encode("utf-8")

    @override
    def loadb(self, data: bytes) -> JSONCompat:
        r: JSONCompat = json.loads(data)
        return r



from __future__ import annotations

import asyncio
import logging
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Generic, TypeVar, cast
from uuid import uuid4

from jobify._internal.common.constants import JobStatus
from jobify._internal.common.datastructures import RequestState, State
from jobify._internal.configuration import Cron
from jobify._internal.context import JobContext, OuterContext
from jobify._internal.exceptions import DuplicateJobError, JobTimeoutError
from jobify._internal.message import (
    AtArguments,
    CronArguments,
    Message,
    PushArguments,
    Triggers,
)
from jobify._internal.scheduler.job import CronContext, Job
from jobify._internal.scheduler.misfire_policy import handle_misfire_policy
from jobify._internal.storage.base import ScheduledJob
from jobify._internal.storage.dummy import DummyStorage

if TYPE_CHECKING:
    from collections.abc import Callable

    from jobify._internal.common.types import ExceptionHandlers
    from jobify._internal.configuration import (
        JobifyConfiguration,
        RouteOptions,
    )
    from jobify._internal.cron_parser import CronParser
    from jobify._internal.inspection import FuncSpec
    from jobify._internal.middleware.base import CallNext, CallNextOuter
    from jobify._internal.runners import Runnable
    from jobify._internal.task_tracker import TaskTracker

WARN_FORCE = (
    "Job {job_id} already scheduled for {schedule}. If you need to "
    "reschedule, cancel the existing job first or use force=True."
)
ReturnT = TypeVar("ReturnT")
logger = logging.getLogger("jobify.scheduler")


class ScheduleBuilder(Generic[ReturnT]):
    __slots__: tuple[str, ...] = (
        "_chain_middleware",
        "_chain_outer_middleware",
        "_configs",
        "_exception_handlers",
        "_runnable",
        "_state",
        "_task_tracker",
        "func_spec",
        "name",
        "route_options",
    )

    def __init__(  # noqa: PLR0913
        self,
        *,
        name: str,
        state: State,
        task_tracker: TaskTracker,
        jobify_config: JobifyConfiguration,
        runnable: Runnable[ReturnT],
        chain_middleware: CallNext,
        chain_outer_middleware: CallNextOuter,
        func_spec: FuncSpec[ReturnT],
        exception_handlers: ExceptionHandlers,
        options: RouteOptions,
    ) -> None:
        self._state: State = state
        self._task_tracker: TaskTracker = task_tracker
        self._configs: JobifyConfiguration = jobify_config
        self._runnable: Runnable[ReturnT] = runnable
        self._chain_middleware: CallNext = chain_middleware
        self._chain_outer_middleware: CallNextOuter = chain_outer_middleware
        self._exception_handlers: ExceptionHandlers = exception_handlers
        self.func_spec: FuncSpec[ReturnT] = func_spec
        self.route_options: RouteOptions = options
        self.name: str = name

    def now(self) -> datetime:
        return datetime.now(tz=self._configs.tz)

    async def cron(
        self,
        cron: str | Cron,
        *,
        job_id: str,
        replace: bool = False,
        force: bool = False,
    ) -> Job[ReturnT]:
        job_id, exists_job = self._ensure_job_id(job_id, replace=replace)
        if isinstance(cron, str):
            cron = Cron(cron)

        real_now = self.now()

        if exists_job is not None:
            job = exists_job
            ctx = cast("CronContext[ReturnT]", exists_job._cron_context)

            if not force and cron == ctx.cron:
                logger.warning(WARN_FORCE.format(job_id=job_id, schedule=cron))
                return exists_job

            old_cron = ctx.cron
            current_offset = ctx.offset
        else:
            job = Job[ReturnT](
                exec_at=real_now,
                job_id=job_id,
                unregister_hook=self._task_tracker.unregister_job,
                storage=self._configs.storage,
            )
            ctx = None
            old_cron = None
            current_offset = real_now

        parser = self._configs.cron_factory(cron.expression)
        offset, next_run_at = self._calculate_schedule_cron(
            old_cron=old_cron,
            new_cron=cron,
            parser=parser,
            offset=current_offset,
            real_now=real_now,
        )
        job.exec_at = next_run_at
        if ctx is not None:
            ctx.cron = cron
            ctx.offset = offset
            ctx.run_count = 0
            ctx.cron_parser = parser
        else:
            ctx = CronContext(
                job=job,
                cron=cron,
                offset=offset,
                run_count=0,
                cron_parser=parser,
            )
            job.bind_cron_context(ctx)
        _ = await self._chain_outer_middleware(
            self._create_outer_context(
                job=job,
                trigger=CronArguments(cron=cron, job_id=job_id, offset=offset),
                is_force=force,
                is_replace=replace,
                schedule_hook=lambda: self._schedule_execution_cron(ctx),
            )
        )
        return job

    async def delay(
        self,
        seconds: float,
        *,
        job_id: str | None = None,
        now: datetime | None = None,
        replace: bool = False,
    ) -> Job[ReturnT]:
        now = now or self.now()
        at = now + timedelta(seconds=seconds)
        return await self.at(at=at, job_id=job_id, replace=replace)

    async def at(
        self,
        at: datetime,
        *,
        job_id: str | None = None,
        replace: bool = False,
        force: bool = False,
    ) -> Job[ReturnT]:
        job_id, exists_job = self._ensure_job_id(job_id, replace=replace)
        if exists_job is not None:
            if not force and exists_job.exec_at == at:
                logger.warning(WARN_FORCE.format(job_id=job_id, schedule=at))
                return exists_job
            exists_job.exec_at = at
            job = exists_job
        else:
            job = Job[ReturnT](
                exec_at=at,
                job_id=job_id,
                storage=self._configs.storage,
                unregister_hook=self._task_tracker.unregister_job,
            )
        _ = await self._chain_outer_middleware(
            self._create_outer_context(
                job=job,
                trigger=AtArguments(job_id=job_id, at=at),
                is_force=force,
                is_replace=replace,
                schedule_hook=lambda: self._schedule_execution_at(job),
            )
        )
        return job

    async def push(self) -> Job[ReturnT]:
        job = Job[ReturnT](
            exec_at=self.now(),
            job_id=uuid4().hex,
            storage=self._configs.storage,
            unregister_hook=self._task_tracker.unregister_job,
        )
        _ = await self._chain_outer_middleware(
            self._create_outer_context(
                job=job,
                trigger=PushArguments(job_id=job.id),
                is_force=False,
                is_replace=False,
                schedule_hook=lambda: self._push_execution(job),
            )
        )
        return job

    def _schedule_execution_cron(
        self,
        ctx: CronContext[ReturnT],
    ) -> asyncio.Handle:
        delay_seconds = self._calculate_delay_seconds(ctx.job.exec_at)
        loop = self._configs.getloop()
        when = loop.time() + delay_seconds
        handle = loop.call_at(when, self._pre_exec_cron, ctx)
        ctx.job.bind_handle(handle)
        return handle

    def _schedule_execution_at(self, job: Job[ReturnT]) -> asyncio.Handle:
        loop = self._configs.getloop()
        delay_seconds = self._calculate_delay_seconds(target_at=job.exec_at)
        if delay_seconds <= 0:
            handle = loop.call_soon(self._pre_exec_at, job)
        else:
            when = loop.time() + delay_seconds
            handle = loop.call_at(when, self._pre_exec_at, job)
        job.bind_handle(handle)
        return handle

    def _push_execution(self, job: Job[ReturnT]) -> asyncio.Handle:
        handle = self._configs.getloop().call_soon(self._pre_exec_at, job)
        job.bind_handle(handle)
        return handle

    async def _exec_job(self, job: Job[ReturnT]) -> None:
        job.status = JobStatus.RUNNING
        job_context = JobContext(
            job=job,
            state=self._state,
            request_state=RequestState(),
            runnable=self._runnable,
            route_options=self.route_options,
            jobify_config=self._configs,
            schedule_builder=self,
        )
        try:
            result = await self._chain_middleware(job_context)
        except JobTimeoutError as exc:
            job.set_exception(exc, status=JobStatus.TIMEOUT)
        except Exception as exc:
            logger.exception("Job %s failed with unexpected error", job.id)
            job.set_exception(exc, status=JobStatus.FAILED)
        else:
            job.set_result(result, status=JobStatus.SUCCESS)

    def _calculate_delay_seconds(self, target_at: datetime) -> float:
        return target_at.timestamp() - self.now().timestamp()

    def _is_persist(self) -> bool:
        is_dummy = isinstance(self._configs.storage, DummyStorage)
        is_durable = self.route_options.get("durable", True)
        return not is_dummy and is_durable

    async def _persist_job(
        self,
        job_id: str,
        next_run_at: datetime,
        trigger: Triggers,
    ) -> None:
        await self._configs.storage.add_schedule(
            self._create_scheduled_job(job_id, next_run_at, trigger)
        )

    def _create_scheduled_job(
        self,
        job_id: str,
        next_run_at: datetime,
        trigger: Triggers,
    ) -> ScheduledJob:
        dumper = self._configs.dumper.dump
        params_type = self.func_spec.type_params
        args_dumped = {
            name: dumper(arg, params_type[name])
            for name, arg in self._runnable.origin_arguments.items()
        }
        msg = Message(
            job_id=job_id,
            name=self.name,
            arguments=args_dumped,
            trigger=trigger,
        )
        serializer = self._configs.serializer.dumpb
        serialized_message = serializer(dumper(msg, Message))
        return ScheduledJob(
            name=self.name,
            job_id=job_id,
            message=serialized_message,
            status=JobStatus.SCHEDULED,
            next_run_at=next_run_at,
        )

    def _ensure_job_id(
        self,
        job_id: str | None,
        *,
        replace: bool,
    ) -> tuple[str, Job[ReturnT] | None]:
        job_id = job_id or uuid4().hex
        if job := self._task_tracker.pending_jobs.get(job_id):
            if replace is True:
                return (job_id, job)

            raise DuplicateJobError(job_id)

        return (job_id, None)

    def _create_outer_context(
        self,
        *,
        job: Job[ReturnT],
        trigger: Triggers,
        is_force: bool,
        is_replace: bool,
        schedule_hook: Callable[[], asyncio.Handle],
    ) -> OuterContext:
        return OuterContext(
            job=job,
            state=self._state,
            trigger=trigger,
            runnable=self._runnable,
            arguments=self._runnable.bound.arguments,
            func_spec=self.func_spec,
            is_force=is_force,
            is_persist=self._is_persist(),
            is_replace=is_replace,
            route_options=self.route_options,
            jobify_config=self._configs,
            request_state=RequestState(),
            persist_job_hook=self._persist_job,
            schedule_hook=schedule_hook,
            schedule_builder=self,
        )

    def _calculate_schedule_cron(
        self,
        *,
        old_cron: Cron | None,
        new_cron: Cron,
        parser: CronParser,
        offset: datetime,
        real_now: datetime,
    ) -> tuple[datetime, datetime]:
        start_date_changed = old_cron is None or (
            old_cron.start_date != new_cron.start_date
        )
        if start_date_changed and new_cron.start_date is not None:
            target_at = new_cron.start_date
            offset = target_at
        else:
            target_at = parser.next_run(now=offset)

        next_run_at = handle_misfire_policy(
            cron_parser=parser,
            next_run_at=target_at,
            real_now=real_now,
            policy=new_cron.misfire_policy,
        )
        return (offset, next_run_at)

    def _pre_exec_at(self, job: Job[ReturnT]) -> None:
        task = asyncio.create_task(self._exec_at(job), name=job.id)
        self._task_tracker.track_task(job.id, task, job._event)

    async def _exec_at(self, job: Job[ReturnT]) -> None:
        await self._exec_job(job)
        if self._is_persist():
            await self._configs.storage.delete_schedule(job.id)
        self._task_tracker.unregister_job(job.id)

    def _pre_exec_cron(self, ctx: CronContext[ReturnT]) -> None:
        task = asyncio.create_task(self._exec_cron(ctx=ctx), name=ctx.job.id)
        self._task_tracker.track_task(ctx.job.id, task, ctx.job._event)

    async def _exec_cron(self, ctx: CronContext[ReturnT]) -> None:
        job = ctx.job
        await self._exec_job(job)

        if job.status is JobStatus.SUCCESS:
            ctx.failure_count = 0
        else:
            ctx.failure_count += 1
        ctx.run_count += 1

        if ctx.is_run_exceeded_by_limit():
            if self._is_persist():
                await self._configs.storage.delete_schedule(job.id)
        elif job.is_reschedulable() and self._configs.app_started:
            if ctx.is_failure_allowed_by_limit():
                offset = ctx.offset = job.exec_at
                next_run_at = ctx.cron_parser.next_run(now=offset)

                if self._is_persist():
                    trigger = CronArguments(
                        job_id=job.id,
                        cron=ctx.cron,
                        offset=offset,
                        run_count=ctx.run_count,
                    )
                    await self._persist_job(job.id, next_run_at, trigger)
                job.update(exec_at=next_run_at, status=JobStatus.SCHEDULED)
                _ = self._schedule_execution_cron(ctx)
                return

            job.status = JobStatus.PERMANENTLY_FAILED
            logger.warning(
                "Job %s stopped due to max failures policy (%s/%s)",
                job.id,
                ctx.failure_count,
                ctx.cron.max_failures,
            )
        self._task_tracker.unregister_job(job.id)

    def _cron(  # noqa: PLR0913
        self,
        *,
        cron: Cron,
        job_id: str,
        offset: datetime,
        next_run_at: datetime,
        cron_parser: CronParser,
        run_count: int,
    ) -> None:
        job = Job[ReturnT](
            exec_at=next_run_at,
            job_id=job_id,
            unregister_hook=self._task_tracker.unregister_job,
            storage=self._configs.storage,
        )
        ctx = CronContext(
            job=job,
            cron=cron,
            offset=offset,
            cron_parser=cron_parser,
            run_count=run_count,
        )
        job.bind_cron_context(ctx)
        self._task_tracker.register_job(job)
        _ = self._schedule_execution_cron(ctx)

    def _at(self, at: datetime, job_id: str) -> None:
        job = Job[ReturnT](
            exec_at=at,
            job_id=job_id,
            unregister_hook=self._task_tracker.unregister_job,
            storage=self._configs.storage,
        )
        self._task_tracker.register_job(job)
        _ = self._schedule_execution_at(job)

    def _push(self, job_id: str, exec_at: datetime) -> None:
        job = Job[ReturnT](
            job_id=job_id,
            exec_at=exec_at,
            storage=self._configs.storage,
            unregister_hook=self._task_tracker.unregister_job,
        )
        self._task_tracker.register_job(job)
        handle = self._configs.getloop().call_soon(self._pre_exec_at, job)
        job.bind_handle(handle)



from __future__ import annotations

import asyncio
import functools
from abc import ABCMeta, abstractmethod
from collections.abc import Awaitable, Callable, Sequence
from typing import (
    Any,
    Protocol,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)

from jobify._internal.context import JobContext, OuterContext

ReturnT = TypeVar("ReturnT")
CallNext: TypeAlias = Callable[[JobContext], Awaitable[Any]]
CallNextOuter: TypeAlias = Callable[[OuterContext], Awaitable[asyncio.Handle]]


@runtime_checkable
class BaseMiddleware(Protocol, metaclass=ABCMeta):
    @abstractmethod
    async def __call__(self, call_next: CallNext, context: JobContext) -> Any:  # noqa: ANN401
        pass


def build_middleware(
    middleware: Sequence[BaseMiddleware],
    /,
    func: Callable[[JobContext], Awaitable[ReturnT]],
) -> CallNext:
    chain_of_middlewares = func
    for m in reversed(middleware):
        chain_of_middlewares = functools.partial(m, chain_of_middlewares)

    return chain_of_middlewares


@runtime_checkable
class BaseOuterMiddleware(Protocol, metaclass=ABCMeta):
    @abstractmethod
    async def __call__(
        self,
        call_next: CallNextOuter,
        context: OuterContext,
    ) -> Any:  # noqa: ANN401
        pass


def build_outer_middleware(
    middleware: Sequence[BaseOuterMiddleware],
    /,
    func: Callable[[OuterContext], Awaitable[asyncio.Handle]],
) -> CallNextOuter:
    chain_of_middlewares = func
    for m in reversed(middleware):
        chain_of_middlewares = functools.partial(m, chain_of_middlewares)

    return chain_of_middlewares



from __future__ import annotations

import inspect
import sys
import uuid
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Generic,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import Unpack, override

from jobify._internal.common.constants import PATCH_FUNC_NAME
from jobify._internal.common.datastructures import State
from jobify._internal.exceptions import RouteAlreadyRegisteredError

if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterator,
        Callable,
        Coroutine,
        Iterator,
        Sequence,
    )

    from jobify._internal.common.types import (
        ExceptionHandlers,
        Lifespan,
        MappingExceptionHandlers,
    )
    from jobify._internal.configuration import RouteOptions
    from jobify._internal.middleware.base import (
        BaseMiddleware,
        BaseOuterMiddleware,
    )
    from jobify._internal.scheduler.job import Job
    from jobify._internal.scheduler.scheduler import ScheduleBuilder


T_co = TypeVar("T_co", covariant=True)
Return_co = TypeVar("Return_co", covariant=True)
ParamsT = ParamSpec("ParamsT")
Route_co = TypeVar("Route_co", bound="Route[..., Any]", covariant=True)
Router_co = TypeVar("Router_co", bound="Router", covariant=True)


def resolve_name(func: Callable[ParamsT, Return_co], /) -> str:
    name = func.__name__
    name = getattr(func, "__qualname__", name).removesuffix(PATCH_FUNC_NAME)
    fmodule = func.__module__
    if name == "<lambda>":
        name = f"lambda_{uuid.uuid4().hex}"
    if fmodule == "__main__":
        mod = inspect.getmodule(func)
        fallback = sys.argv[0] if sys.argv else "."
        fmodule = Path(getattr(mod, "__file__", fallback)).stem
    return f"{fmodule}:{name}"


class Route(ABC, Generic[ParamsT, Return_co]):
    def __init__(
        self,
        name: str,
        func: Callable[ParamsT, Return_co],
        options: RouteOptions,
    ) -> None:
        self.name: str = name
        self.func: Callable[ParamsT, Return_co] = func
        self.options: RouteOptions = options

    def __call__(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> Return_co:
        return self.func(*args, **kwargs)

    @overload
    def schedule(
        self: Route[ParamsT, Coroutine[object, object, T_co]],
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ScheduleBuilder[T_co]: ...

    @overload
    def schedule(
        self: Route[ParamsT, Return_co],
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ScheduleBuilder[Return_co]: ...

    @abstractmethod
    def schedule(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ScheduleBuilder[Any]:
        raise NotImplementedError

    @overload
    async def push(
        self: Route[ParamsT, Coroutine[object, object, T_co]],
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> Job[T_co]: ...

    @overload
    async def push(
        self: Route[ParamsT, Return_co],
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> Job[Return_co]: ...

    @abstractmethod
    async def push(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> Job[Any]:
        raise NotImplementedError


class Registrator(ABC, Generic[Route_co]):
    def __init__(  # noqa: PLR0913
        self,
        *,
        state: State | None,
        lifespan: Lifespan[Router_co] | None,
        route_class: type[Route_co],
        middleware: Sequence[BaseMiddleware] | None,
        outer_middleware: Sequence[BaseOuterMiddleware] | None,
        exception_handlers: MappingExceptionHandlers | None,
    ) -> None:
        self._routes: dict[str, Route_co] = {}
        self._middleware: list[BaseMiddleware] = list(middleware or [])
        self._outer_middleware: list[BaseOuterMiddleware] = list(
            outer_middleware or []
        )
        self._state_lifespan: Final = self._iter_lifespan(
            lifespan or dummy_lifespan
        )
        self._exception_handlers: ExceptionHandlers = dict(
            exception_handlers or {}
        )
        self.state: State = state or State()
        self.route_class: type[Route_co] = route_class

    @overload
    def __call__(
        self,
        func: Callable[ParamsT, Return_co],
    ) -> Route[ParamsT, Return_co]: ...

    @overload
    def __call__(
        self,
        **options: Unpack[RouteOptions],
    ) -> Callable[
        [Callable[ParamsT, Return_co]], Route[ParamsT, Return_co]
    ]: ...

    @overload
    def __call__(
        self,
        func: Callable[ParamsT, Return_co],
        **options: Unpack[RouteOptions],
    ) -> Route[ParamsT, Return_co]: ...

    def __call__(
        self,
        func: Callable[ParamsT, Return_co] | None = None,
        **options: Unpack[RouteOptions],
    ) -> (
        Route[ParamsT, Return_co]
        | Callable[[Callable[ParamsT, Return_co]], Route[ParamsT, Return_co]]
    ):
        wrapper = self._register(options)
        if callable(func):
            return wrapper(func)
        return wrapper  # pragma: no cover

    def _register(
        self,
        options: RouteOptions,
    ) -> Callable[[Callable[ParamsT, Return_co]], Route[ParamsT, Return_co]]:
        def wrapper(
            func: Callable[ParamsT, Return_co],
        ) -> Route[ParamsT, Return_co]:
            if isinstance(func, Route):
                func = cast("Callable[ParamsT, Return_co]", func.func)

            name = options.get("name") or resolve_name(func)
            if name in self._routes:
                raise RouteAlreadyRegisteredError(name)

            return self.register(name, func, options)

        return wrapper

    @abstractmethod
    def register(
        self,
        name: str,
        func: Callable[ParamsT, Return_co],
        options: RouteOptions,
    ) -> Route[ParamsT, Return_co]:
        raise NotImplementedError

    async def _iter_lifespan(
        self,
        user_lifespan: Lifespan[Any],
    ) -> AsyncIterator[None]:
        async with user_lifespan(self) as maybe_state:
            if maybe_state is not None:
                self.state.update(maybe_state)
            yield None

    async def emit_startup(self) -> None:
        await anext(self._state_lifespan)

    async def emit_shutdown(self) -> None:
        await anext(self._state_lifespan, None)


@asynccontextmanager
async def dummy_lifespan(_: Router) -> AsyncIterator[None]:
    yield None


class Router(ABC):
    def __init__(self, *, prefix: str | None) -> None:
        self.prefix: str = prefix if prefix else ""
        self._parent: Router | None = None
        self._sub_routers: list[Router] = []

    @property
    @abstractmethod
    def task(self) -> Registrator[Route[..., Any]]:
        raise NotImplementedError

    @override
    def __repr__(self) -> str:
        return f"<{type(self).__name__}>"

    @property
    def chain_tail(self) -> Iterator[Router]:
        yield self
        for router in self.sub_routers:
            yield from router.chain_tail

    @property
    def routes(self) -> Iterator[Route[..., Any]]:
        yield from self.task._routes.values()

    @property
    def sub_routers(self) -> Sequence[Router]:
        return self._sub_routers

    @property
    def parent(self) -> Router | None:
        return self._parent

    @parent.setter
    def parent(self, router: Router) -> None:
        """Set the parent router for this router (internal use only).

        Do not use this method in own code.
        All routers should be included via the `include_router` method.
        Self- and circular-referencing are not allowed here.
        """
        if self._parent:
            msg = f"Router is already attached to {self._parent!r}"
            raise RuntimeError(msg)
        if self is router:
            msg = "Self-referencing routers is not allowed"
            raise RuntimeError(msg)

        parent: Router | None = router

        while parent is not None:
            if parent is self:
                msg = "Circular referencing of Router is not allowed"
                raise RuntimeError(msg)
            parent = parent.parent

        self._parent = router

    def include_router(self, router: Router) -> None:
        router.parent = self
        self._sub_routers.append(router)

    def include_routers(self, *routers: Router) -> None:
        if not routers:
            msg = "At least one router must be provided"
            raise ValueError(msg)
        for router in routers:
            self.include_router(router)

    def add_middleware(self, middleware: BaseMiddleware) -> None:
        self.task._middleware.append(middleware)

    def add_outer_middleware(self, middleware: BaseOuterMiddleware) -> None:
        self.task._outer_middleware.append(middleware)

    def remove_route(self, name: str) -> None:
        del self.task._routes[name]

    def add_route(self, route: Route[..., Any]) -> None:
        self.task._routes[route.name] = route



from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from datetime import datetime
from typing import Protocol, TypeAlias, runtime_checkable

Expression: TypeAlias = str
CronFactory: TypeAlias = Callable[[Expression], "CronParser"]


@runtime_checkable
class CronParser(Protocol, metaclass=ABCMeta):
    @abstractmethod
    def next_run(self, *, now: datetime) -> datetime:
        raise NotImplementedError



from datetime import datetime
from typing import Any, NamedTuple, TypeAlias

from jobify._internal.configuration import Cron

Triggers: TypeAlias = "PushArguments | AtArguments | CronArguments"


class PushArguments(NamedTuple):
    job_id: str


class AtArguments(NamedTuple):
    job_id: str
    at: datetime


class CronArguments(NamedTuple):
    job_id: str
    cron: Cron
    offset: datetime
    run_count: int = 0


class Message(NamedTuple):
    job_id: str
    name: str
    arguments: dict[str, Any]
    trigger: Triggers



from __future__ import annotations

import base64
import dataclasses
import json
from collections.abc import Callable, Iterable, Sequence
from datetime import datetime, timedelta
from decimal import Decimal
from enum import Enum
from typing import Any, ClassVar, Protocol, TypeAlias, get_args, get_type_hints
from zoneinfo import ZoneInfo

from typing_extensions import TypeIs, override

from jobify._internal.serializers.base import JSONCompat, Serializer


class DataclassType(Protocol):
    __dataclass_fields__: ClassVar[dict[str, dataclasses.Field[Any]]]


class NamedTupleType(Protocol):
    _asdict: ClassVar[Callable[[NamedTupleType], dict[str, SupportedTypes]]]


SupportedTypes: TypeAlias = (
    None
    | int
    | str
    | bool
    | Enum
    | float
    | bytes
    | Decimal
    | ZoneInfo
    | datetime
    | timedelta
    | DataclassType
    | NamedTupleType
    | set["SupportedTypes"]
    | list["SupportedTypes"]
    | tuple["SupportedTypes", ...]
    | dict[str, "SupportedTypes"]
)
TypeRegistry: TypeAlias = dict[str, Callable[..., SupportedTypes]]


def is_named_tuple_type(tp: Any) -> bool:  # noqa: ANN401
    return (
        isinstance(tp, type)
        and issubclass(tp, tuple)
        and hasattr(tp, "_fields")  # pyright: ignore[reportUnknownArgumentType]
    )


def is_named_tuple(o: SupportedTypes) -> TypeIs[NamedTupleType]:
    return isinstance(o, tuple) and hasattr(o, "_asdict")


def is_dataclass(o: SupportedTypes) -> TypeIs[DataclassType]:
    return dataclasses.is_dataclass(o) and not isinstance(o, type)


def is_structured_type(tp: Any) -> bool:  # noqa: ANN401
    return (
        dataclasses.is_dataclass(tp)
        or is_named_tuple_type(tp)
        or (
            hasattr(tp, "__origin__")
            and dataclasses.is_dataclass(tp.__origin__)
        )
    )


def json_extended_encoder(o: SupportedTypes) -> JSONCompat:  # noqa: C901, PLR0911
    if is_dataclass(o):  # pragma: no cover
        return {
            "__dataclass__": {
                "type": o.__class__.__name__,
                "fields": {
                    f.name: json_extended_encoder(getattr(o, f.name))
                    for f in dataclasses.fields(o)
                },
            }
        }
    if is_named_tuple(o):
        return {
            "__namedtuple__": {
                "type": o.__class__.__name__,
                "fields": {
                    k: json_extended_encoder(v) for k, v in o._asdict().items()
                },
            }
        }
    if isinstance(o, Enum):
        return {
            "__enum__": {
                "type": o.__class__.__name__,
                "value": o.value,
            }
        }
    if isinstance(o, datetime):
        return {"__datetime__": o.isoformat()}
    if isinstance(o, Decimal):
        return {"__decimal__": str(o)}
    if isinstance(o, tuple):
        return {"__tuple__": [json_extended_encoder(item) for item in o]}
    if isinstance(o, set):
        return {"__set__": [json_extended_encoder(item) for item in o]}
    if isinstance(o, list):
        return [json_extended_encoder(item) for item in o]
    if isinstance(o, dict):
        return {k: json_extended_encoder(v) for k, v in o.items()}
    if isinstance(o, bytes):
        return {"__bytes__": base64.b64encode(o).decode("utf-8")}
    if isinstance(o, timedelta):
        return {"__timedelta__": o.total_seconds()}
    if isinstance(o, ZoneInfo):
        return {"__zoneinfo__": o.key}
    return o


class JsonDecoderHook:
    def __init__(self, registry: TypeRegistry) -> None:
        self.registry: TypeRegistry = registry

    def __call__(self, dct: dict[str, Any]) -> SupportedTypes:  # noqa: C901, PLR0911
        if "__dataclass__" in dct:
            data = dct["__dataclass__"]
            fields = data["fields"]
            type_name = data["type"]
            return self.registry[type_name](**fields)

        if "__namedtuple__" in dct:
            data = dct["__namedtuple__"]
            fields = data["fields"]
            type_name = data["type"]
            return self.registry[type_name](**fields)

        if "__enum__" in dct:
            data = dct["__enum__"]
            type_name = data["type"]
            return self.registry[type_name](data["value"])

        if "__datetime__" in dct:
            return datetime.fromisoformat(dct["__datetime__"])
        if "__decimal__" in dct:
            return Decimal(dct["__decimal__"])
        if "__bytes__" in dct:
            return base64.b64decode(dct["__bytes__"])
        if "__tuple__" in dct:
            return tuple(dct["__tuple__"])
        if "__set__" in dct:
            return set(dct["__set__"])
        if "__timedelta__" in dct:
            return timedelta(seconds=dct["__timedelta__"])
        if "__zoneinfo__" in dct:
            return ZoneInfo(dct["__zoneinfo__"])
        return dct


class ExtendedJSONSerializer(Serializer):
    def __init__(
        self,
        registry: Sequence[Callable[..., SupportedTypes]] = (),
    ) -> None:
        self.registry: TypeRegistry = {}
        self.add_system_types(registry)
        self.decoder_hook: JsonDecoderHook = JsonDecoderHook(self.registry)

    def add_system_types(
        self, tp: Sequence[Callable[..., SupportedTypes]], /
    ) -> None:
        self.registry.update({t.__name__: t for t in tp})

    @override
    def dumpb(self, data: SupportedTypes) -> bytes:
        json_compat = json_extended_encoder(data)
        return json.dumps(json_compat).encode("utf-8")

    @override
    def loadb(self, data: bytes) -> SupportedTypes:
        hook = self.decoder_hook
        decoded: SupportedTypes = json.loads(data, object_hook=hook)
        return decoded

    def register_hints(self, types: Iterable[Any]) -> None:
        for tp in types:
            if not is_structured_type(tp):
                continue
            if getattr(tp, "__name__", None) == "JobContext":
                continue
            if args := get_args(tp):
                self.register_hints(args)
                continue
            if tp.__name__ in self.registry:
                continue

            self.registry[tp.__name__] = tp
            field_hints = get_type_hints(tp)
            self.register_hints(field_hints.values())



# pyright: reportImportCycles=false
import asyncio
from collections.abc import Callable, Mapping
from contextlib import AbstractAsyncContextManager
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar

if TYPE_CHECKING:
    from jobify._internal.context import JobContext

AppType = TypeVar("AppType")

StatelessLifespan: TypeAlias = Callable[
    [AppType],
    AbstractAsyncContextManager[None],
]
StatefulLifespan: TypeAlias = Callable[
    [AppType],
    AbstractAsyncContextManager[Mapping[str, Any]],
]
Lifespan: TypeAlias = StatelessLifespan[AppType] | StatefulLifespan[AppType]
LoopFactory: TypeAlias = Callable[[], asyncio.AbstractEventLoop]


ExceptionHandler: TypeAlias = Callable[[Exception, "JobContext"], Any]
ExceptionHandlers: TypeAlias = dict[type[Exception], ExceptionHandler]
MappingExceptionHandlers: TypeAlias = Mapping[
    type[Exception], ExceptionHandler
]



"""Package provides various storage solutions for Jobify."""

from jobify._internal.storage.base import Storage
from jobify._internal.storage.sqlite import SQLiteStorage

__all__ = ("SQLiteStorage", "Storage")



import multiprocessing
import sys
from collections.abc import Collection, Mapping
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, TypedDict, final
from zoneinfo import ZoneInfo

from jobify._internal.common.constants import INFINITY, RunMode
from jobify._internal.common.types import (
    LoopFactory,
    MappingExceptionHandlers,
)
from jobify._internal.cron_parser import CronFactory
from jobify._internal.scheduler.misfire_policy import (
    GracePolicy,
    MisfirePolicy,
)
from jobify._internal.serializers.base import Serializer
from jobify._internal.storage.base import Storage
from jobify._internal.typeadapter.base import Dumper, Loader


@final
class WorkerPools:
    __slots__: tuple[str, ...] = ("_processpool", "threadpool")

    def __init__(
        self,
        *,
        _processpool: ProcessPoolExecutor | None = None,
        threadpool: ThreadPoolExecutor | None = None,
    ) -> None:
        self._processpool = _processpool
        self.threadpool = threadpool

    @property
    def processpool(self) -> ProcessPoolExecutor:  # pragma: no cover
        if self._processpool is None:
            if sys.platform in ("win32", "darwin"):
                start_method = "spawn"
            elif "forkserver" in multiprocessing.get_all_start_methods():
                start_method = "forkserver"
            else:
                start_method = "spawn"
            mp_ctx = multiprocessing.get_context(start_method)
            self._processpool = ProcessPoolExecutor(mp_context=mp_ctx)
        return self._processpool

    def close(self) -> None:
        if self._processpool is not None:
            self._processpool.shutdown(wait=True, cancel_futures=True)
            self._processpool = None


@final
class JobifyConfiguration:
    __slots__: tuple[str, ...] = (
        "app_started",
        "cron_factory",
        "dumper",
        "getloop",
        "loader",
        "serializer",
        "storage",
        "tz",
        "worker_pools",
    )

    def __init__(  # noqa: PLR0913
        self,
        *,
        tz: ZoneInfo,
        dumper: Dumper,
        loader: Loader,
        storage: Storage,
        getloop: LoopFactory,
        serializer: Serializer,
        worker_pools: WorkerPools,
        cron_factory: CronFactory,
        app_started: bool = False,
    ) -> None:
        self.tz = tz
        self.dumper = dumper
        self.loader = loader
        self.storage = storage
        self.getloop = getloop
        self.serializer = serializer
        self.worker_pools = worker_pools
        self.cron_factory = cron_factory
        self.app_started = app_started


@dataclass(slots=True, kw_only=True, order=True)
class Cron:
    expression: str = field(kw_only=False)
    max_runs: int = INFINITY
    max_failures: int = 10
    misfire_policy: MisfirePolicy | GracePolicy = MisfirePolicy.ONCE
    start_date: datetime | None = None
    args: Collection[Any] = ()
    kwargs: Mapping[str, Any] = field(default_factory=dict)  # pyright: ignore[reportUnknownVariableType]

    def __post_init__(self) -> None:
        if self.max_failures < 1:
            msg = (
                "max_cron_failures must be >= 1."
                " Use 1 for 'stop on first error'."
            )
            raise ValueError(msg)


class RouteOptions(TypedDict, total=False):
    name: str
    cron: Cron | str
    retry: int
    timeout: float
    durable: bool
    run_mode: RunMode
    metadata: Mapping[str, Any]
    exception_handlers: MappingExceptionHandlers



from __future__ import annotations

import functools
import logging
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    ParamSpec,
    TypeAlias,
    TypeVar,
    cast,
    get_type_hints,
)

from typing_extensions import override

from jobify._internal.common.constants import (
    PATCH_CRON_DEF_ID,
    PATCH_FUNC_NAME,
    JobStatus,
)
from jobify._internal.configuration import Cron
from jobify._internal.context import inject_context
from jobify._internal.exceptions import (
    raise_app_already_started_error,
    raise_app_not_started_error,
)
from jobify._internal.inspection import FuncSpec, make_func_spec
from jobify._internal.middleware.base import (
    BaseOuterMiddleware,
    CallNextOuter,
    build_middleware,
    build_outer_middleware,
)
from jobify._internal.middleware.exceptions import ExceptionMiddleware
from jobify._internal.middleware.retry import RetryMiddleware
from jobify._internal.middleware.timeout import TimeoutMiddleware
from jobify._internal.router.base import Registrator, Route, Router
from jobify._internal.runners import Runnable, create_run_strategy
from jobify._internal.scheduler.scheduler import ScheduleBuilder
from jobify._internal.serializers.json_extended import ExtendedJSONSerializer

if TYPE_CHECKING:
    import asyncio
    import inspect
    from collections.abc import Callable, Iterator, Sequence

    from jobify._internal.common.datastructures import State
    from jobify._internal.common.types import (
        ExceptionHandler,
        ExceptionHandlers,
        Lifespan,
        MappingExceptionHandlers,
    )
    from jobify._internal.configuration import (
        JobifyConfiguration,
        RouteOptions,
    )
    from jobify._internal.context import JobContext, OuterContext
    from jobify._internal.middleware.base import BaseMiddleware, CallNext
    from jobify._internal.router.node import NodeRouter
    from jobify._internal.runners import RunStrategy
    from jobify._internal.scheduler.job import Job
    from jobify._internal.task_tracker import TaskTracker


logger = logging.getLogger("Jobify.router")

ReturnT = TypeVar("ReturnT")
ParamsT = ParamSpec("ParamsT")
RootRouter_co = TypeVar("RootRouter_co", bound="RootRouter", covariant=True)

CRONS_DEF_KEY = "__crons_definition__"
CronsDefinition: TypeAlias = dict[str, tuple["RootRoute[..., Any]", Cron]]


class RootRoute(Route[ParamsT, ReturnT]):
    def __init__(  # noqa: PLR0913
        self,
        *,
        name: str,
        func: Callable[ParamsT, ReturnT],
        func_spec: FuncSpec[ReturnT],
        state: State,
        options: RouteOptions,
        strategy: RunStrategy[ParamsT, ReturnT],
        task_tracker: TaskTracker,
        exception_handlers: ExceptionHandlers,
        jobify_config: JobifyConfiguration,
    ) -> None:
        super().__init__(name, func, options)
        self._run_strategy: RunStrategy[ParamsT, ReturnT] = strategy
        self._task_state: TaskTracker = task_tracker
        self._chain_middleware: CallNext | None = None
        self._chain_outer_middleware: CallNextOuter | None = None
        self._exception_handlers: ExceptionHandlers = exception_handlers
        self.state: State = state
        self.func_spec: FuncSpec[ReturnT] = func_spec
        self.jobify_config: JobifyConfiguration = jobify_config

        # --------------------------------------------------------------------
        # HACK: ProcessPoolExecutor / Multiprocessing  # noqa: ERA001, FIX004
        #
        # Problem: `ProcessPoolExecutor` (used for ExecutionMode.PROCESS)
        # serializes the function by its name. When we use `@register`
        # as a decorator, the function's name in the module (`my_func`)
        # now points to the `FuncWrapper` object, not the original function.
        # This breaks `pickle`.
        #
        # Solution: We rename the *original* function (adding a suffix)
        # and "inject" it back into its own module under this new
        # name. This way, `ProcessPoolExecutor` can find and pickle it.
        #
        # We DO NOT apply this hack in two cases (Guard Clauses):
        # 1. If `register` is used as a direct function call (`reg(my_func)`),
        #    because `my_func` in the module still points to the original.
        # 2. If the function has already been renamed (protects from re-entry).
        # --------------------------------------------------------------------

        # Guard 1: Protect against double-renaming
        if func.__name__.endswith(PATCH_FUNC_NAME):
            return

        # Guard 2: Check if `register` is used as a decorator (@)
        # or as a direct function call.
        module = sys.modules[func.__module__]
        module_attr = getattr(module, func.__name__, None)
        if module_attr is func:
            return

        # Apply the hack: rename and inject back into the module
        new_name = f"{func.__name__}{PATCH_FUNC_NAME}"
        func.__name__ = new_name
        if hasattr(func, "__qualname__"):  # pragma: no cover
            original_qualname = func.__qualname__.rsplit(".", 1)
            original_qualname[-1] = new_name
            new_qualname = ".".join(original_qualname)
            func.__qualname__ = new_qualname
        setattr(module, new_name, func)

    @override
    def schedule(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> ScheduleBuilder[ReturnT]:
        bound = self.func_spec.signature.bind(*args, **kwargs)
        return self.create_builder(bound)

    @override
    async def push(
        self,
        *args: ParamsT.args,
        **kwargs: ParamsT.kwargs,
    ) -> Job[ReturnT]:
        bound = self.func_spec.signature.bind(*args, **kwargs)
        return await self.create_builder(bound).push()

    def create_builder(
        self,
        bound: inspect.BoundArguments,
        /,
    ) -> ScheduleBuilder[ReturnT]:
        if (
            self.jobify_config.app_started is False
            or self._chain_middleware is None
            or self._chain_outer_middleware is None
        ):
            raise_app_not_started_error("schedule")
        return ScheduleBuilder(
            name=self.name,
            state=self.state,
            options=self.options,
            func_spec=self.func_spec,
            task_tracker=self._task_state,
            jobify_config=self.jobify_config,
            chain_middleware=self._chain_middleware,
            exception_handlers=self._exception_handlers,
            chain_outer_middleware=self._chain_outer_middleware,
            runnable=Runnable(
                name=self.name,
                bound=bound,
                strategy=self._run_strategy,
                func_spec=self.func_spec,
            ),
        )


class RootRegistrator(Registrator[RootRoute[..., Any]]):
    def __init__(  # noqa: PLR0913
        self,
        *,
        state: State | None,
        lifespan: Lifespan[RootRouter_co] | None,
        middleware: Sequence[BaseMiddleware] | None,
        outer_middleware: Sequence[BaseOuterMiddleware] | None,
        task_tracker: TaskTracker,
        jobify_config: JobifyConfiguration,
        exception_handlers: MappingExceptionHandlers | None,
        route_class: type[RootRoute[..., Any]],
    ) -> None:
        super().__init__(
            state=state,
            lifespan=lifespan,
            route_class=route_class,
            middleware=middleware,
            outer_middleware=outer_middleware,
            exception_handlers=exception_handlers,
        )
        self._task_tracker: TaskTracker = task_tracker
        self._jobify_config: JobifyConfiguration = jobify_config
        self._system_middleware: tuple[BaseMiddleware, ...] = (
            RetryMiddleware(),
            ExceptionMiddleware(jobify_config),
            TimeoutMiddleware(),
        )

    @override
    def register(
        self,
        name: str,
        func: Callable[ParamsT, ReturnT],
        options: RouteOptions,
    ) -> RootRoute[ParamsT, ReturnT]:
        if self._jobify_config.app_started is True:
            raise_app_already_started_error("register")

        if isinstance(self._jobify_config.serializer, ExtendedJSONSerializer):
            hints = get_type_hints(func)
            self._jobify_config.serializer.register_hints(hints.values())

        strategy = create_run_strategy(
            func,
            self._jobify_config,
            mode=options.get("run_mode"),
        )
        route = self.route_class(
            name=name,
            func=func,
            func_spec=make_func_spec(func),
            state=self.state,
            options=options,
            strategy=strategy,
            task_tracker=self._task_tracker,
            jobify_config=self._jobify_config,
            exception_handlers=self._exception_handlers,
        )
        _ = functools.update_wrapper(route, func)
        self._routes[name] = route

        if cron := options.get("cron"):
            if isinstance(cron, str):
                cron = Cron(cron)

            job_id = f"{route.name}{PATCH_CRON_DEF_ID}"
            p = {job_id: (route, cron)}
            self.state.setdefault(CRONS_DEF_KEY, {}).update(p)

        return cast("RootRoute[ParamsT, ReturnT]", route)


class RootRouter(Router):
    def __init__(  # noqa: PLR0913
        self,
        *,
        state: State | None,
        lifespan: Lifespan[RootRouter_co] | None,
        middleware: Sequence[BaseMiddleware] | None,
        outer_middleware: Sequence[BaseOuterMiddleware] | None,
        task_tracker: TaskTracker,
        jobify_config: JobifyConfiguration,
        exception_handlers: MappingExceptionHandlers | None,
        route_class: type[RootRoute[..., Any]],
    ) -> None:
        super().__init__(prefix=None)
        self._registrator: RootRegistrator = RootRegistrator(
            state=state,
            lifespan=lifespan,
            middleware=middleware,
            outer_middleware=outer_middleware,
            task_tracker=task_tracker,
            jobify_config=jobify_config,
            exception_handlers=exception_handlers,
            route_class=route_class,
        )
        self.state: State = self._registrator.state
        self.exception_handlers: ExceptionHandlers = (
            self._registrator._exception_handlers
        )

    @property
    @override
    def task(self) -> RootRegistrator:
        return self._registrator

    def add_exception_handler(
        self,
        cls_exc: type[Exception],
        handler: ExceptionHandler,
    ) -> None:
        if self.task._jobify_config.app_started is True:
            raise_app_already_started_error("add_exception_handler")
        self.task._exception_handlers[cls_exc] = handler

    @override
    def add_middleware(self, middleware: BaseMiddleware) -> None:
        if self.task._jobify_config.app_started is True:
            raise_app_already_started_error("add_middleware")
        super().add_middleware(middleware)

    @override
    def add_outer_middleware(self, middleware: BaseOuterMiddleware) -> None:
        if self.task._jobify_config.app_started is True:
            raise_app_already_started_error("add_outer_middleware")
        super().add_outer_middleware(middleware)

    @override
    def include_router(self, router: Router) -> None:
        super().include_router(router)

    def propagate_real_routes(self) -> None:
        for router in self.sub_routers:
            self._propagate_real_routes(cast("NodeRouter", router))

    def _propagate_real_routes(self, router: NodeRouter) -> None:
        for route in tuple(router.routes):
            router.remove_route(route.name)
            prefix = f"{router.prefix}:" if router.prefix else ""
            route.name = f"{prefix}{route.name}"
            real_route = self.task.register(
                route.name,
                route.func,
                route.options,
            )
            route.bind(real_route)
            router.add_route(real_route)

        for sub_router in router.sub_routers:
            suffix = f".{sub_router.prefix}" if sub_router.prefix else ""
            sub_router.prefix = f"{router.prefix}{suffix}"
            self._propagate_real_routes(sub_router)

    async def _propagate_startup(self, router: Router) -> None:
        await router.task.emit_startup()

        merged_middleware = [
            *router.task._middleware,
            *self.task._system_middleware,
        ]
        outer_middleware = router.task._outer_middleware
        chain = build_middleware(merged_middleware, self._execution)
        chain_outer = build_outer_middleware(outer_middleware, self._schedule)

        for route in cast("Iterator[RootRoute[..., Any]]", router.routes):
            route.state = router.task.state
            route._chain_middleware = chain
            route._chain_outer_middleware = chain_outer
            route._exception_handlers = {
                **router.task._exception_handlers,
                **route.options.get("exception_handlers", {}),
            }

        for sub_router in router.sub_routers:
            sub_router.task.state = router.task.state | sub_router.task.state
            sub_router.task._exception_handlers = (
                router.task._exception_handlers
                | sub_router.task._exception_handlers
            )
            sub_router.task._middleware = [
                *router.task._middleware,
                *sub_router.task._middleware,
            ]
            sub_router.task._outer_middleware = [
                *router.task._outer_middleware,
                *sub_router.task._outer_middleware,
            ]
            await self._propagate_startup(sub_router)

    async def _propagate_shutdown(self) -> None:
        for router in self.chain_tail:
            await router.task.emit_shutdown()

    async def _schedule(self, context: OuterContext) -> asyncio.Handle:
        if context.is_replace and (handle := context.job._handle) is not None:
            handle.cancel()
        if context.is_persist:
            await context.persist_job_hook(
                context.job.id,
                context.job.exec_at,
                context.trigger,
            )
        self.task._task_tracker.register_job(context.job)
        handle = context.schedule_hook()
        context.job.status = JobStatus.SCHEDULED
        return handle

    async def _execution(self, context: JobContext) -> Any:  # noqa: ANN401
        inject_context(context)
        return await context.runnable()
